"""
This module is a modified version of PointNet++ py-torch based on https://github.com/yanx27/Pointnet_Pointnet2_pytorch
"""
from typing import Tuple, Union

import torch
import torch.nn.functional as F
from einops import repeat  # type: ignore
from torch import Tensor
from torch.nn import BatchNorm1d, Conv1d, Dropout, Linear, Module, ReLU, Sequential

from pycarus.learning.models.pointnet import TNet
from pycarus.learning.models.pointnet2_utils import PointNetFeaturePropagation
from pycarus.learning.models.pointnet2_utils import PointNetSetAbstraction
from pycarus.learning.models.pointnet2_utils import PointNetSetAbstractionMsg


class PointNet2EncoderMSG(Module):
    """Model class respresenting the PointNet++ Encoder network as proposed in:

    "Qi, C. R., Yi, L., Su, H., & Guibas, L. J. (2017).
    Pointnet++: Deep hierarchical feature learning on point sets in a metric space.
    Advances in Neural Information Processing Systems."
    """

    def __init__(
        self,
        channels_in: int = 3,
        size_global_feature: int = 1024,
        use_tnet: bool = True,
        align_normals: bool = False,
    ) -> None:
        """Create a new instance of PointNet++ Encoder.

        Args:
            size_global_feature (optional): the size for the global feature. Defaults to 1024.
            use_tnet (optional): if True use the points T-Net. Defaults to False.
            use_normals (optional): if True aspect as input also point normals. Defaults to False.
        """
        super(PointNet2EncoderMSG, self).__init__()

        self.align_normals = align_normals
        self.use_tnet = use_tnet
        self.size_global_feature = size_global_feature
        self.num_channels = channels_in - 3

        if self.use_tnet:
            self.tnet = TNet(channels_in=channels_in, dimension=3)

        self.sa_1 = PointNetSetAbstractionMsg(
            512,
            [0.1, 0.2, 0.4],
            [16, 32, 128],
            self.num_channels,
            [[32, 32, 64], [64, 64, 128], [64, 96, 128]],
        )
        self.sa_2 = PointNetSetAbstractionMsg(
            128,
            [0.2, 0.4, 0.8],
            [32, 64, 128],
            320,
            [[64, 64, 128], [128, 128, 256], [128, 128, 256]],
        )
        self.sa_3 = PointNetSetAbstraction(
            0,
            0,
            0,
            640 + 3,
            [256, 512, self.size_global_feature],
            True,
        )

    def forward(self, x: Tensor) -> Tuple[Tensor, Tensor, Union[Tensor, None]]:
        """Process input point cloud and produce point net++ features.

        Args:
            x: The input point cloud(s) with shape (B, NUM_POINTS, D + features).

        Raises:
            ValueError: If use normals is True but only point coordinates are provided.

        Returns:
            - The computed global feature vector with shape (B, SIZE_GLOBAL_FEAT).
            - The sampled point at the last layer (B, NUM_PTS).
            - The learned transformation matrix for the input points.
        """

        if self.use_tnet:
            x = torch.transpose(x, 2, 1)
            trans_pts = self.tnet(x)
            x = self.tnet.transform(x, trans_pts, self.align_normals, False)
            x = torch.transpose(x, 2, 1)
        else:
            trans_pts = None

        xyz = x[:, :, :3]
        feats = x[:, :, 3:] if x.shape[-1] > 3 else None

        xyz_l1, feats_l1 = self.sa_1(xyz, feats)
        xyz_l2, feats_l2 = self.sa_2(xyz_l1, feats_l1)
        xyz_l3, feats_l3 = self.sa_3(xyz_l2, feats_l2)
        global_feature = feats_l3.squeeze(1).reshape(-1, self.size_global_feature)

        return global_feature, xyz_l3, trans_pts


class PointNet2ClassificationMSG(Module):
    def __init__(
        self,
        channels_in: int = 3,
        size_global_feature: int = 1024,
        use_tnet: bool = False,
        align_normals: bool = False,
        dropout_p: float = 0.5,
        num_classes: int = 10,
    ) -> None:
        """Create a new instance of PointNet++ Classification Network.

        Args:
            channels_in (optional): the size of first convolutional layer. Defaults to 3.
            size_global_feature (optional): the size for the global feature. Defaults to 1024.
            use_tnet (optional): if True use the points T-Net. Defaults to False.
            align_normals (optional): if True rotate normals using the T-Net. Defaults to False.
            dropout_p (optional): the dropout probability. Defaults to 0.5.
            num_classes (optional): the number of classes. Defaults to 10.
        """

        super(PointNet2ClassificationMSG, self).__init__()
        self.encoder = PointNet2EncoderMSG(
            channels_in=channels_in,
            size_global_feature=size_global_feature,
            use_tnet=use_tnet,
            align_normals=align_normals,
        )

        self.fc_layers = Sequential(
            Linear(size_global_feature, 512, bias=False),
            BatchNorm1d(512),  # type: ignore
            ReLU(True),
            Linear(512, 256, bias=False),
            BatchNorm1d(256),  # type: ignore
            ReLU(True),
            Dropout(dropout_p),
            Linear(256, num_classes),
        )

    def forward(self, x: torch.Tensor) -> Tuple[Tensor, Union[Tensor, None]]:
        """Process input point clouds and produces claffication scores.

        Args:
            x: The input clouds with shape (B, NUM_PTS, DIM). The method internally
            transposes the data tensor to correctly operate with PyTorch convetion (B CH N).

        Returns:
            A tuple containing:
                - The computed classification scores with shape (B, NUM_CLASSES).
                - The learned transformation matrix for the input points.
        """
        global_feature, _, trans = self.encoder(x)
        return self.fc_layers(global_feature), trans


class PointNet2PartSegmentationMSG(Module):
    def __init__(
        self,
        channels_in: int = 3,
        size_global_feature: int = 1024,
        use_tnet: bool = True,
        align_normals: bool = False,
        dropout_p: float = 0.5,
        num_classes: int = 16,
        num_classes_part: int = 50,
    ) -> None:
        """Create a new instance of PointNet++ Part Segmentation Network.

        Args:
            channels_in (optional): the size of first convolutional layer. Defaults to 3.
            size_global_feature (optional): the size for the global feature. Defaults to 2048.
            use_pts_tnet (optional): if True use the points T-Net. Defaults to True.
            align_normals (optional): if True rotate normals using the T-Net. Defaults to True.
            dropout_p (_type_, optional): _description_. Defaults to 0.5.
            num_classes (optional): the number of classes for the categorical vector. Defaults to 16.
            num_classes_part: (optional): the number of overal part segmentation classes Defaults to 50.
        """
        super(PointNet2PartSegmentationMSG, self).__init__()

        additional_channels = channels_in - 3
        self.align_normals = align_normals
        self.channels_in = channels_in
        self.use_tnet = use_tnet
        if self.use_tnet:
            self.tnet = TNet(channels_in=channels_in, dimension=3)

        self.sa_1 = PointNetSetAbstractionMsg(
            512,
            [0.1, 0.2, 0.4],
            [32, 64, 128],
            3 + additional_channels,
            [[32, 32, 64], [64, 64, 128], [64, 96, 128]],
        )

        self.sa_2 = PointNetSetAbstractionMsg(
            128,
            [0.4, 0.8],
            [64, 128],
            128 + 128 + 64,
            [[128, 128, 256], [128, 196, 256]],
        )

        self.sa_3 = PointNetSetAbstraction(
            num_point=0,
            radius=0,
            num_samples=0,
            in_channel=512 + 3,
            mlp=[256, 512, size_global_feature],
            group_all=True,
        )

        self.fp_1 = PointNetFeaturePropagation(in_channel=150 + additional_channels, mlp=[128, 128])
        self.fp_2 = PointNetFeaturePropagation(in_channel=576, mlp=[256, 128])
        self.fp_3 = PointNetFeaturePropagation(in_channel=size_global_feature + 512, mlp=[256, 256])

        self.num_classes = num_classes
        self.conv_1 = Conv1d(128, 128, 1)
        self.bn_1 = BatchNorm1d(128)  # type: ignore
        self.drop_1 = Dropout(dropout_p)
        self.conv_2 = Conv1d(128, num_classes_part, 1)

    def forward(self, x: Tensor, vector_cat: Tensor) -> Tuple[Tensor, Union[Tensor, None]]:
        """Process input point clouds and produces per points claffication scores.

        Args:
            x: The input clouds with shape (B, NUM_PTS, DIM).
            vector_cat: the one hot encoding categorical vector (B, NUM_CLASSES).

        Returns:
            A tuple containing:
                - The computed classification scores with shape (B, NUM_PTS, NUM_CLASSES_PART).
                - The learned transformation matrix for the input points.
        """
        _, num_points, _ = x.shape

        if self.use_tnet:
            x = torch.transpose(x, 2, 1)
            trans = self.tnet(x)
            x = self.tnet.transform(x, trans, self.align_normals, False)
            x = torch.transpose(x, 2, 1)
        else:
            trans = None

        if self.channels_in > 3:
            feats_l0 = x
            xyz = x[:, :, :3]
        else:
            feats_l0 = x
            xyz = x

        xyz_l1, feats_l1 = self.sa_1(xyz, feats_l0)
        xyz_l2, feats_l2 = self.sa_2(xyz_l1, feats_l1)
        xyz_l3, feats_l3 = self.sa_3(xyz_l2, feats_l2)

        # Feature Propagation layers
        feats_l2 = self.fp_3(xyz_l2, xyz_l3, feats_l2, feats_l3)
        feats_l1 = self.fp_2(xyz_l1, xyz_l2, feats_l1, feats_l2)

        one_hot_enc = repeat(vector_cat, "b n -> b r n", r=num_points)

        feats_l0 = self.fp_1(
            xyz,
            xyz_l1,
            torch.cat([one_hot_enc, xyz, feats_l0], -1),
            feats_l1,
        )

        feats_l0 = torch.transpose(feats_l0, 2, 1)
        feats_out = F.relu(self.bn_1(self.conv_1(feats_l0)))
        feats_out = self.drop_1(feats_out)
        feats_out = self.conv_2(feats_out)
        feats_out = feats_out.permute(0, 2, 1)

        return feats_out, trans



from pathlib import Path
from typing import Callable, List, Tuple

import h5py  # type: ignore
import torch
from torch import Tensor
from torch.utils.data import Dataset

from pycarus.transforms.var import Compose


class ModelNet40(Dataset):
    def __init__(
        self,
        root: Path,
        split: str,
        transforms: List[Callable] = [],
    ) -> None:
        """Create ModelNet40 dataset.

        Splits are obtained according to the paper:
        "Revisiting Point Cloud Shape Classification with a Simple and Effective Baseline",
        Ankit Goyal, Hei Law, Bowei Liu, Alejandro Newell, Jia Deng.

        Args:
            root: The directory that contains the dataset.
            split: The required split. Available splits are:
                - "train",  the training set without the validation split.
                - "test", the standard test set.
                - "val", the validation split.
                - "train_and_val", the training set with the validation split.
            transforms: The list of transforms to apply to pcds. Defaults to [].
        """
        super().__init__()

        assert split in ["train_and_val", "val", "train", "test"]
        self.split = split

        self.dataset_root = root

        h5_files = [self.dataset_root / f for f in self.get_h5_files_from_split(self.split)]
        h5_files.sort()

        self.pcds, self.labels = self.load_h5(h5_files)

        self.transform = Compose(transforms)

        with open(self.dataset_root / "shape_names.txt") as f:
            self.shape_names = [line.strip() for line in f.readlines()]

    def get_h5_files_from_split(self, split: str) -> List[str]:
        """Get the h5 files to be loaded according to the given split.

        Args:
            split: The required split.

        Returns:
            The list of h5 files to be loaded.
        """
        files = {
            "train_and_val": ["train_no_val.h5", "val.h5"],
            "train": ["train_no_val.h5"],
            "val": ["val.h5"],
            "test": ["test_0.h5", "test_1.h5"],
        }
        return files[split]

    @staticmethod
    def load_h5(paths: List[Path]) -> Tuple[Tensor, Tensor]:
        """Load all the given h5 files.

        Args:
            paths: The list of paths to the h5 files to load.

        Returns:
            - One tensor with all the pcds, with shape (NUM_PCDS, NUM_POINTS, 3).
            - One tensor with all the labels, with shape (NUM_PCDS,).
        """
        all_pcds: List[Tensor] = []
        all_labels: List[Tensor] = []
        for h5_path in paths:
            f = h5py.File(h5_path, "r+")
            pcds = f["data"][:].astype("float32")
            labels = f["label"][:].astype("int64")
            f.close()
            all_pcds.append(torch.from_numpy(pcds))
            all_labels.append(torch.from_numpy(labels))
        return torch.cat(all_pcds, dim=0), torch.cat(all_labels, dim=0).squeeze(1)

    def __len__(self) -> int:
        """Return the number of pcds in the dataset.

        Returns:
            The number of pcds in the dataset.
        """
        return self.pcds.shape[0]

    def __getitem__(self, index: int) -> Tuple[Tensor, Tensor]:
        """Return the pcd and the label at the given index.

        Args:
            index: The index of the required element.

        Returns:
            - A tensor with the point cloud with shape (NUM_POINTS, 3).
            - A tensor with the label.
        """
        pcd, label = self.pcds[index], self.labels[index]

        pcd = self.transform(pcd)

        return pcd, label

    def get_class_name(self, label: int) -> str:
        return self.shape_names[label]



from pathlib import Path
from typing import Callable, List, Tuple

import h5py  # type: ignore
import numpy as np
import torch
import torch.utils.data as data
from torch import Tensor

from pycarus.transforms.var import Compose

T_ITEM = Tuple[str, str, Tensor, Tensor]


class Mvp(data.Dataset):
    CATEGORIES = [
        "02691156",
        "02933112",
        "02958343",
        "03001627",
        "03636649",
        "04256520",
        "04379243",
        "04530566",
        "02818832",
        "02828884",
        "02871439",
        "02924116",
        "03467517",
        "03790512",
        "03948459",
        "04225987",
    ]

    def __init__(
        self,
        root: Path,
        split: str,
        num_points: int = 2048,
        categories: List[str] = [],
        transforms_complete: List[Callable] = [],
        transforms_incomplete: List[Callable] = [],
        transforms_all: List[Callable] = [],
    ) -> None:
        """Class implementing the MVP dataset as proposed in:

        Pan, L., Chen, X., Cai, Z., Zhang, J., Zhao, H., Yi, S., & Liu, Z. (2021).
        Variational Relational Point Completion Network.
        In Proceedings of the IEEE/CVF Conference on Computer Vision and Pattern Recognition (pp. 8524-8533).

        Args:
            root: The path to the folder containing the dataset.
            split: The name of the split to load. Allowed values: train, val, train_and_val, test.
            num_points: The resolution for groundtruth clouds. Allowed values: 2048, 4096, 8192, 16384.
            categories: A list of categories ids to select. Defaults to [] (keep all categories).
            transforms_complete: The transforms to apply to the complete clouds. Defaults to [].
            transforms_incomplete: The transforms to apply to the incomplete clouds. Defaults to [].
            transforms_all: The transforms to apply to both the complete
                and the incomplete clouds. Defaults to [].
        """
        splits = ["train", "val", "train_and_val", "test"]
        if split not in splits:
            raise ValueError(f"Split {split} not allowed, only {splits} are allowed.")

        self.num_points = num_points
        res_gt = [2048, 4096, 8192, 16384]
        if self.num_points not in res_gt:
            raise ValueError(f"{self.num_points} value not allowed, only {res_gt} are allowed.")

        self.transform_complete = Compose(transforms_complete)
        self.transform_incomplete = Compose(transforms_incomplete)
        self.transform_all = Compose(transforms_all)

        split_file = split if split not in ["val", "train_and_val"] else "train"
        self.path_inc = root / f"mvp_{split_file}_input.h5"
        file_inc = h5py.File(self.path_inc, "r")
        self.incompletes = np.array(file_inc.get("incomplete_pcds"))
        self.labels = np.array(file_inc.get("labels"))
        incompletes_novel = np.array(file_inc.get("novel_incomplete_pcds"))
        labels_novel = np.array(file_inc.get("novel_labels"))
        file_inc.close()

        self.path_comp = root / f"mvp_{split_file}_gt_{self.num_points}pts.h5"
        file_comp = h5py.File(self.path_comp, "r")
        self.completes = np.array(file_comp.get("complete_pcds"))
        completes_novel = np.array(file_comp.get("novel_complete_pcds"))
        file_comp.close()

        self.incompletes = np.concatenate((self.incompletes, incompletes_novel), axis=0)
        self.completes = np.concatenate((self.completes, completes_novel), axis=0)
        self.labels = np.concatenate((self.labels, labels_novel), axis=0)

        indices = list(range(self.incompletes.shape[0]))

        if split != "test":
            with open(root / "val_shapes.txt") as f:
                val_shapes = [int(line.strip()) for line in f.readlines()]

            val_indices: List[int] = []
            for idx in val_shapes:
                start = idx * 26
                end = start + 26
                val_indices.extend(range(start, end))

            train_indices = [idx for idx in indices if idx not in val_indices]

            if split == "train":
                indices = train_indices
            elif split == "val":
                indices = val_indices
            else:
                indices = sorted(train_indices + val_indices)

        if len(categories) == 0:
            self.indices = list(range(self.incompletes.shape[0]))
        else:
            self.indices = []

            for index in range(self.incompletes.shape[0]):
                label = int(self.labels[index])
                if Mvp.CATEGORIES[label] in categories:
                    self.indices.append(index)

    def __len__(self) -> int:
        """Return the number of items in the dataset.

        Returns:
            The number of items in the dataset.
        """
        return len(self.indices)

    def __getitem__(self, index: int) -> T_ITEM:
        """Return one item of the dataset.

        Args:
            index: The index of the required item.

        Returns:
            - The category of the item.
            - The name of the item formatted as "s{shape_index}_v{view_index}".
            - A tensor with the incomplete point cloud with shape (NUM_POINTS, 3).
            - A tensor with the complete point cloud with shape (NUM_POINTS, 3).
        """
        idx = self.indices[index]
        idx_complete = idx // 26
        idx_view = idx % 26

        incomplete = torch.from_numpy((self.incompletes[idx]))
        complete = torch.from_numpy((self.completes[idx_complete]))

        incomplete = self.transform_all(self.transform_incomplete(incomplete))
        complete = self.transform_all(self.transform_complete(complete))

        label = torch.tensor(self.labels[idx], dtype=torch.long)
        category = Mvp.CATEGORIES[label]
        name = f"s{idx_complete}_v{idx_view:02d}"

        return category, name, incomplete, complete



from typing import Tuple, Union

import torch
import torch.nn.functional as F
from einops import rearrange, repeat  # type: ignore
from pytorch3d.ops import knn_gather, knn_points  # type: ignore
from torch import Tensor
from torch.nn import BatchNorm1d, BatchNorm2d, Conv1d, Conv2d, Dropout, LeakyReLU, Linear, Module
from torch.nn import Sequential


class TNet(Module):
    """Model class respresenting the T-Net as proposed in:

    Wang, Yue, et al. "Dynamic graph cnn for learning on point clouds."
    Acm Transactions On Graphics (tog) 38.5 (2019): 1-12.
    """

    def __init__(self) -> None:
        super(TNet, self).__init__()

        self.conv_1 = Sequential(
            Conv2d(6, 64, kernel_size=1, bias=False),
            BatchNorm2d(64),  # type: ignore
            LeakyReLU(negative_slope=0.2),
        )
        self.conv_2 = Sequential(
            Conv2d(64, 128, kernel_size=1, bias=False),
            BatchNorm2d(128),  # type: ignore
            LeakyReLU(negative_slope=0.2),
        )
        self.conv_3 = Sequential(
            Conv1d(128, 1024, kernel_size=1, bias=False),
            BatchNorm1d(1024),  # type: ignore
            LeakyReLU(negative_slope=0.2),
        )

        self.fc_1 = Sequential(
            Linear(1024, 512, bias=False),
            BatchNorm1d(512),  # type: ignore
            LeakyReLU(negative_slope=0.2),
        )

        self.fc_2 = Sequential(
            Linear(512, 256, bias=False),
            BatchNorm1d(256),  # type: ignore
            LeakyReLU(negative_slope=0.2),
        )

        self.fc_3 = Linear(256, 9)

    def forward(self, x: Tensor) -> Tensor:
        """Process input point cloud and produce the transformation matrix.

        Args:
            x: The input clouds with shape (B, 6, NUM_PTS, K).

        Returns:
            The learned transformation with shape (B, 3, 3).
        """

        x = self.conv_1(x)
        x = self.conv_2(x)
        x = x.max(dim=-1, keepdim=False)[0]

        x = self.conv_3(x)
        x = x.max(dim=-1, keepdim=False)[0]

        x = self.fc_1(x)
        x = self.fc_2(x)
        x = self.fc_3(x)

        iden = torch.eye(3).reshape(1, 9)
        iden = iden.to(x.device)
        iden = repeat(iden, "b d -> (r b)  d", r=x.size(0))
        x = x + iden
        x = x.reshape(-1, 3, 3)

        return x


def get_graph_feature(x: Tensor, k: int = 20, indices: Union[Tensor, None] = None) -> Tensor:
    """Select features from neighbors.

    Args:
        x : the input features with shape (B, NUM_PTS, D).
        k: the number of neighbors. Defaults to 20.
        indices: the features indices with shape (B, NUM_PTS, K). Defaults to None.

    Returns:
        The selected features with shape (B, NUM_PTS, K, 2*D)
    """
    if indices is None:
        _, indices, _ = knn_points(x, x, K=k)

    features = knn_gather(x, indices)

    x = repeat(x, "b n d -> b n r d", r=k)
    features = torch.cat((features - x, x), dim=3)

    return features


class DGCNNEncoder(Module):
    """Model class respresenting the DGCNN Encoder network as proposed in:

    Wang, Yue, et al. "Dynamic graph cnn for learning on point clouds."
    Acm Transactions On Graphics (tog) 38.5 (2019): 1-12.
    """

    def __init__(
        self,
        k: int = 20,
        size_global_feature: int = 1024,
        use_geometry_nn: bool = False,
        use_tnet: bool = False,
    ) -> None:
        """Create a new instance of DGCNN Classification Encoder.

        Args:
            k (optional): the number of k-nearest neighbors for the Edge-Conv. Defaults to 20.
            size_global_feature (optional): the size for the global feature. Defaults to 1024.
            use_geometry_nn (optional): if True compute the nearest neighbors in coordinate space
            use_tnet (optional): if True use the points T-Net. Defaults to False.
        """
        super(DGCNNEncoder, self).__init__()

        self.k = k
        self.size_feat_glob = size_global_feature
        self.use_use_geometry_nn = use_geometry_nn
        self.use_pts_tnet = use_tnet
        if self.use_pts_tnet:
            self.tnet = TNet()

        self.bn_1 = BatchNorm1d(64)  # type: ignore
        self.bn_2 = BatchNorm1d(64)  # type: ignore
        self.bn_3 = BatchNorm1d(128)  # type: ignore
        self.bn_4 = BatchNorm1d(256)  # type: ignore
        self.bn_5 = BatchNorm1d(self.size_feat_glob)  # type: ignore

        self.slope = 0.2
        self.fc_1 = Linear(3 * 2, 64, bias=False)
        self.fc_2 = Linear(64 * 2, 64, bias=False)
        self.fc_3 = Linear(64 * 2, 128, bias=False)
        self.fc_4 = Linear(128 * 2, 256, bias=False)
        self.fc_5 = Linear(512, self.size_feat_glob, bias=False)

    def block_forward(
        self, x: Tensor, linear: Module, bn: Module, indices: Union[Tensor, None] = None
    ) -> Tensor:
        batch_size, num_points = x.shape[0], x.shape[1]

        x = get_graph_feature(x, k=self.k, indices=indices)
        x = linear(x)
        x = rearrange(x, "b n k d -> b d (n k)")
        x = bn(x)
        x = F.leaky_relu(x, negative_slope=self.slope)
        x = x.reshape(batch_size, -1, num_points, self.k)

        features_out = x.max(dim=-1, keepdim=False)[0]
        features_out = rearrange(features_out, "b ch n -> b n ch")
        return features_out

    def forward(self, x: Tensor) -> Tuple[Tensor, Union[Tensor, None]]:
        """Process input point cloud and produce dgcnn features.

        Args:
            x: The input clouds with shape (B, NUM_PTS, DIM).

        Returns:
            A tuple containing:
                - The computed global feature vector with shape (B, SIZE_GLOBAL_FEAT).
                - The learned transformation matrix for the input points.
        """
        if self.use_pts_tnet:
            x0 = get_graph_feature(x, k=self.k)
            x0 = torch.permute(x0, (0, 3, 1, 2))
            trans = self.tnet(x0)
            x = torch.bmm(x, trans)
        else:
            trans = None

        indices = None
        if self.use_use_geometry_nn:
            _, indices, _ = knn_points(x, x, K=self.k)

        x_1 = self.block_forward(x, self.fc_1, self.bn_1, indices)
        x_2 = self.block_forward(x_1, self.fc_2, self.bn_2, indices)
        x_3 = self.block_forward(x_2, self.fc_3, self.bn_3, indices)
        x_4 = self.block_forward(x_3, self.fc_4, self.bn_4, indices)
        x_5 = torch.cat((x_1, x_2, x_3, x_4), dim=-1)

        x_5 = self.fc_5(x_5)
        x_5 = rearrange(x_5, "b n d -> b d n")
        x_5 = self.bn_5(x_5)
        feat = F.leaky_relu(x_5, negative_slope=self.slope)

        global_feature = F.adaptive_max_pool1d(feat, 1)
        global_feature = global_feature.reshape(len(x), self.size_feat_glob)

        return global_feature, trans


class DGCNNClassification(Module):
    """Model class respresenting the DGCNN Classification network as proposed in:

    Wang, Yue, et al. "Dynamic graph cnn for learning on point clouds."
    Acm Transactions On Graphics (tog) 38.5 (2019): 1-12.
    """

    def __init__(
        self,
        k: int = 20,
        size_global_feature: int = 1024,
        use_geometry_nn: bool = False,
        use_tnet: bool = False,
        dropout_p: float = 0.5,
        num_classes: int = 10,
    ) -> None:
        """Create a new instance of DGCNN Classification Network.

        Args:
            k (optional): the number of k-nearest neighbors for the Edge-Conv. Defaults to 20.
            size_global_feature (optional): the size for the global feature. Defaults to 1024.
            use_geometry_nn (optional): if True compute the nearest neighbors in coordinate space
            instead of feature space.
            use_tnet (optional): if True use the points T-Net. Defaults to False.
            dropout_p (optional): the dropout probability. Defaults to 0.5.
            num_classes (optional): the number of classes. Defaults to 10.
        """

        super(DGCNNClassification, self).__init__()
        self.encoder = DGCNNEncoder(
            use_tnet=use_tnet,
            k=k,
            size_global_feature=size_global_feature,
            use_geometry_nn=use_geometry_nn,
        )
        self.fc_1 = Linear(1024, 512)
        self.fc_2 = Linear(512, 256)
        self.fc_3 = Linear(256, num_classes)
        self.dropout = Dropout(dropout_p)
        self.bn_1 = BatchNorm1d(512)  # type: ignore
        self.bn_2 = BatchNorm1d(256)  # type: ignore

    def forward(self, x: Tensor) -> Tuple[Tensor, Union[Tensor, None]]:
        """Process input point clouds and produces claffication scores.

        Args:
            x: The input clouds with shape (B, NUM_PTS, DIM).

        Returns:
            A tuple containing:
                - The computed classification scores with shape (B, NUM_CLASSES).
                - The learned transformation matrix for the input points.
        """
        x, trans_pts = self.encoder(x)
        x = F.relu(self.dropout(self.bn_1(self.fc_1(x))))
        x = F.relu(self.dropout(self.bn_2(self.fc_2(x))))
        x = self.fc_3(x)
        return x, trans_pts


class DGCNNPartSegmentation(Module):
    """Model class respresenting the DGCNN Part Segmentation network as proposed in:

    Wang, Yue, et al. "Dynamic graph cnn for learning on point clouds."
    Acm Transactions On Graphics (tog) 38.5 (2019): 1-12.
    """

    def __init__(
        self,
        k: int = 20,
        size_global_feature: int = 1024,
        use_geometry_nn: bool = False,
        use_tnet: bool = True,
        dropout_p: float = 0.5,
        num_classes: int = 16,
        num_classes_part: int = 10,
    ) -> None:
        """Create a new instance of DGCNN Part Segmentation Network.

        Args:
            k (optional): the number of k-nearest neighbors for the Edge-Conv. Defaults to 20.
            size_global_feature (optional): the size for the global feature. Defaults to 1024.
            use_geometry_nn (optional): if True compute the nearest neighbors in coordinate space
            instead of feature space.
            use_tnet (optional): if True use the points T-Net. Defaults to False.
            dropout_p (optional): the dropout probability. Defaults to 0.5.
            num_classes (optional): the number of classes for the categorical vector. Defaults to 10.
            num_classes_part: (optional): the number of overal part segmentation classes Defaults to 10.
        """
        super(DGCNNPartSegmentation, self).__init__()

        self.num_classes_part = num_classes_part
        self.k = k
        self.use_pts_tnet = use_tnet
        self.use_geometry_nn = use_geometry_nn
        self.slope = 0.2

        if self.use_pts_tnet:
            self.tnet = TNet()

        self.conv_1 = Sequential(
            Conv2d(6, 64, kernel_size=1, bias=False),
            BatchNorm2d(64),  # type: ignore
            LeakyReLU(negative_slope=0.2),
        )
        self.conv_2 = Sequential(
            Conv2d(64, 64, kernel_size=1, bias=False),
            BatchNorm2d(64),  # type: ignore
            LeakyReLU(negative_slope=0.2),
        )
        self.conv_3 = Sequential(
            Conv2d(64 * 2, 64, kernel_size=1, bias=False),
            BatchNorm2d(64),  # type: ignore
            LeakyReLU(negative_slope=0.2),
        )
        self.conv_4 = Sequential(
            Conv2d(64, 64, kernel_size=1, bias=False),
            BatchNorm2d(64),  # type: ignore
            LeakyReLU(negative_slope=0.2),
        )
        self.conv_5 = Sequential(
            Conv2d(64 * 2, 64, kernel_size=1, bias=False),
            BatchNorm2d(64),  # type: ignore
            LeakyReLU(negative_slope=0.2),
        )
        self.conv_6 = Sequential(
            Conv1d(192, size_global_feature, kernel_size=1, bias=False),
            BatchNorm1d(size_global_feature),  # type: ignore
            LeakyReLU(negative_slope=0.2),
        )
        self.conv_7 = Sequential(
            Conv1d(num_classes, 64, kernel_size=1, bias=False),
            BatchNorm1d(64),  # type: ignore
            LeakyReLU(negative_slope=0.2),
        )

        size_in_conv_8 = 64 + size_global_feature + 64 + 64 + 64
        self.conv_8 = Sequential(
            Conv1d(size_in_conv_8, 256, kernel_size=1, bias=False),
            BatchNorm1d(256),  # type: ignore
            LeakyReLU(negative_slope=0.2),
        )

        self.dropout = Dropout(dropout_p)
        self.conv_9 = Sequential(
            Conv1d(256, 256, kernel_size=1, bias=False),
            BatchNorm1d(256),  # type: ignore
            LeakyReLU(negative_slope=0.2),
        )

        self.conv_10 = Sequential(
            Conv1d(256, 128, kernel_size=1, bias=False),
            BatchNorm1d(128),  # type: ignore
            LeakyReLU(negative_slope=0.2),
        )
        self.conv_11 = Conv1d(128, num_classes_part, kernel_size=1, bias=False)

    def forward(self, x: Tensor, vector_cat: Tensor) -> Tuple[Tensor, Union[Tensor, None]]:
        """Process input point clouds and produces per points claffication scores.

        Args:
            x: The input clouds with shape (B, NUM_PTS, DIM).
            vector_cat: the one hot encoding categorical vector (B, NUM_CLASSES).

        Returns:
            A tuple containing:
                - The computed classification scores with shape (B, NUM_PTS, NUM_CLASSES_PART).
                - The learned transformation matrix for the input points.
        """
        batch_size = x.size(0)
        num_points = x.size(1)

        if self.use_pts_tnet:
            x0 = get_graph_feature(x, k=self.k)
            x0 = torch.permute(x0, (0, 3, 1, 2))
            trans = self.tnet(x0)
            x = torch.bmm(x, trans)
        else:
            trans = None

        indices = None
        if self.use_geometry_nn:
            _, indices, _ = knn_points(x, x, K=self.k)

        x = get_graph_feature(x, k=self.k, indices=indices)
        x = torch.permute(x, (0, 3, 1, 2))

        x = self.conv_1(x)
        x = self.conv_2(x)
        x_1 = x.max(dim=-1, keepdim=False)[0]

        x = torch.transpose(x_1, 2, 1)
        x = get_graph_feature(x, k=self.k, indices=indices)
        x = torch.permute(x, (0, 3, 1, 2))
        x = self.conv_3(x)
        x = self.conv_4(x)
        x_2 = x.max(dim=-1, keepdim=False)[0]

        x = torch.transpose(x_2, 2, 1)
        x = get_graph_feature(x, k=self.k, indices=indices)
        x = torch.permute(x, (0, 3, 1, 2))
        x = self.conv_5(x)
        x_3 = x.max(dim=-1, keepdim=False)[0]

        x = torch.cat((x_1, x_2, x_3), dim=1)

        x = self.conv_6(x)
        x = x.max(dim=-1, keepdim=True)[0]

        vector_cat = vector_cat.reshape(batch_size, -1, 1)
        vector_cat = self.conv_7(vector_cat)

        x = torch.cat((x, vector_cat), dim=1)
        x = repeat(x, "b ch n -> b ch (n r)", r=num_points)

        x = torch.cat((x, x_1, x_2, x_3), dim=1)

        x = self.conv_8(x)
        x = self.dropout(x)
        x = self.conv_9(x)
        x = self.dropout(x)
        x = self.conv_10(x)
        x = self.conv_11(x)
        x = torch.transpose(x, 2, 1)

        return x, trans



from typing import Tuple, Union

import torch
import torch.nn.functional as F
from einops import repeat  # type: ignore
from torch import Tensor
from torch.nn import BatchNorm1d, Conv1d, Dropout, Linear, Module


class TNet(Module):
    """Model class respresenting the T-Net as proposed in:

    Qi, Charles R., et al.
    "Pointnet: Deep learning on point sets for 3d classification and segmentation."
    Proceedings of the IEEE conference on computer vision and pattern recognition. 2017.

    """

    def __init__(self, channels_in: int = 3, dimension: int = 3) -> None:
        """Create a new instance of T-Net.

        Args:
            channels_in (optional): the size of first convolutional layer. Defaults to 3.
            dimension (optional): the dimension for the learned transformation. Defaults to 3.
        """
        super(TNet, self).__init__()
        self.dim = dimension
        self.conv_1 = Conv1d(channels_in, 64, 1)
        self.conv_2 = Conv1d(64, 128, 1)
        self.conv_3 = Conv1d(128, 1024, 1)
        self.fc_1 = Linear(1024, 512)
        self.fc_2 = Linear(512, 256)
        self.fc_3 = Linear(256, self.dim * self.dim)

        self.bn_1 = BatchNorm1d(64)  # type: ignore
        self.bn_2 = BatchNorm1d(128)  # type: ignore
        self.bn_3 = BatchNorm1d(1024)  # type: ignore
        self.bn_4 = BatchNorm1d(512)  # type: ignore
        self.bn_5 = BatchNorm1d(256)  # type: ignore

    def forward(self, x: Tensor) -> Tensor:
        """Process input point cloud and produce the transformation matrix.

        Args:
            x: The input clouds with shape (B, CH_IN, NUM_PTS). The method internally
            does NOT transposes the data tensor to correctly operate with PyTorch convetion (B CH N).

        Returns:
            The learned transformation with shape (B, DIMENSION, DIMENSION).
        """
        batch_size = x.shape[0]
        x = F.relu(self.bn_1(self.conv_1(x)))
        x = F.relu(self.bn_2(self.conv_2(x)))
        x = F.relu(self.bn_3(self.conv_3(x)))
        x = torch.max(x, 2, keepdim=True)[0]
        x = x.reshape(batch_size, -1)

        x = F.relu(self.bn_4(self.fc_1(x)))
        x = F.relu(self.bn_5(self.fc_2(x)))
        x = self.fc_3(x)

        iden = torch.eye(self.dim).reshape(1, self.dim * self.dim)
        iden = iden.to(x.device)
        iden = repeat(iden, "b d -> (r b)  d", r=x.size(0))
        x = x + iden
        x = x.reshape(-1, self.dim, self.dim)

        return x

    def transform(
        self,
        x: Tensor,
        transform: Tensor,
        rotate_normals: bool,
        apply_to_features: bool,
    ) -> Tensor:
        """Apply the transformation.

        Args:
            x: the input points to transform with shape (B, D, N).
            transform: the transformation to apply with shape (B, D, D).

        Returns:
            The transformed points with shape (B, D, N).
        """

        def multiply(x: Tensor, transform: Tensor) -> Tensor:
            x = x.transpose(2, 1)
            x = torch.bmm(x, transform)
            x = x.transpose(2, 1)
            return x

        if apply_to_features or x.shape[1] == 3:
            return multiply(x, transform)
        else:
            pts = x[:, :3, :]
            features = x[:, 3:, :]
            pts = multiply(pts, transform)

            if rotate_normals:
                features = multiply(features, transform)

            x = torch.cat([pts, features], dim=1)
            return x


class PointNetEncoder(Module):
    """Model class respresenting the PointNet Encoder as proposed in:

    Qi, Charles R., et al.
    "Pointnet: Deep learning on point sets for 3d classification and segmentation."
    Proceedings of the IEEE conference on computer vision and pattern recognition. 2017.
    """

    def __init__(
        self,
        channels_in: int = 3,
        size_global_feature: int = 1024,
        use_pts_tnet: bool = True,
        use_feats_tnet: bool = True,
        align_normals: bool = False,
    ) -> None:
        """Create a new instance of PointNet Encoder.

        Args:
            channels_in (optional): the size of first convolutional layer. Defaults to 3.
            size_global_feature (optional): the size for the global feature. Defaults to 1024.
            use_pts_tnet (optional): If True use the points T-Net. Defaults to True.
            use_feats_tnet (optional): If True use the features T-Net. Defaults to True.
            align_normals (optional): If True rotate normals using the T-Net. Defaults to True.
        """
        super(PointNetEncoder, self).__init__()
        self.size_feat_glob = size_global_feature
        self.use_pts_tnet = use_pts_tnet
        self.use_feat_tnet = use_feats_tnet
        self.align_normals = align_normals

        if self.use_pts_tnet:
            self.tnet_pts = TNet(channels_in=channels_in, dimension=3)

        self.conv_1 = Conv1d(channels_in, 64, 1)
        self.conv_2 = Conv1d(64, 128, 1)
        self.conv_3 = Conv1d(128, self.size_feat_glob, 1)
        self.bn_1 = BatchNorm1d(64)  # type: ignore
        self.bn_2 = BatchNorm1d(128)  # type: ignore
        self.bn_3 = BatchNorm1d(self.size_feat_glob)  # type: ignore

        if self.use_feat_tnet:
            self.tnet_feats = TNet(channels_in=64, dimension=64)

    def forward(self, x: Tensor) -> Tuple[Tensor, Union[Tensor, None], Union[Tensor, None]]:
        """Process input point cloud and produce point net features.

        Args:
            x: The input clouds with shape (B, NUM_PTS, DIM). The method internally
            transposes the data tensor to correctly operate with PyTorch convetion (B CH N).

        Returns:
            A tuple containing:
                - The computed global feature vector with shape (B, SIZE_GLOBAL_FEAT).
                - The learned transformation matrix for the input points.
                - The learned transformation matrix for the features.
        """
        x = x.transpose(2, 1)  # B DIM NUM_PTS
        if self.use_pts_tnet:
            trans_pts = self.tnet_pts(x)
            x = self.tnet_pts.transform(x, trans_pts, self.align_normals, False)
        else:
            trans_pts = None

        x = F.relu(self.bn_1(self.conv_1(x)))

        if self.use_feat_tnet:
            trans_feat = self.tnet_feats(x)
            x = self.tnet_feats.transform(x, trans_feat, False, True)
        else:
            trans_feat = None

        x = F.relu(self.bn_2(self.conv_2(x)))
        x = self.bn_3(self.conv_3(x))
        x = torch.max(x, 2, keepdim=True)[0]
        global_feature = x.reshape(-1, self.size_feat_glob)

        return global_feature, trans_pts, trans_feat


class PointNetClassification(Module):
    def __init__(
        self,
        channels_in: int = 3,
        size_global_feature: int = 1024,
        use_pts_tnet: bool = True,
        use_feats_tnet: bool = True,
        align_normals: bool = False,
        dropout_p: float = 0.3,
        num_classes: int = 10,
    ) -> None:
        """Create a new instance of PointNet Classification Network.

        Args:
            channels_in (optional): the size of first convolutional layer. Defaults to 3.
            size_global_feature (optional): the size for the global feature. Defaults to 1024.
            use_pts_tnet (optional): If True use the points T-Net. Defaults to True.
            use_feats_tnet (optional): If True use the features T-Net. Defaults to True.
            align_normals (optional): If True rotate normals using the T-Net. Defaults to False.
            dropout_p: the dropout probability. Defaults to 0.5.
            num_classes: the number of classes. Defaults to 10.
        """

        super(PointNetClassification, self).__init__()
        self.encoder = PointNetEncoder(
            channels_in=channels_in,
            size_global_feature=size_global_feature,
            align_normals=align_normals,
            use_pts_tnet=use_pts_tnet,
            use_feats_tnet=use_feats_tnet,
        )
        self.fc_1 = Linear(1024, 512)
        self.fc_2 = Linear(512, 256)
        self.fc_3 = Linear(256, num_classes)
        self.dropout = Dropout(dropout_p)
        self.bn_1 = BatchNorm1d(512)  # type: ignore
        self.bn_2 = BatchNorm1d(256)  # type: ignore

    def forward(self, x: Tensor) -> Tuple[Tensor, Union[Tensor, None], Union[Tensor, None]]:
        """Process input point clouds and produces claffication scores.

        Args:
            x: The input clouds with shape (B, NUM_PTS, DIM). The method internally
            transposes the data tensor to correctly operate with PyTorch convetion (B CH N).

        Returns:
            A tuple containing:
                - The computed classification scores with shape (B, NUM_CLASSES).
                - The learned transformation matrix for the input points.
                - The learned transformation matrix for the features.
        """
        x, trans_pts, trans_feat = self.encoder(x)
        x = F.relu(self.dropout(self.bn_1(self.fc_1(x))))
        x = F.relu(self.dropout(self.bn_2(self.fc_2(x))))
        x = self.fc_3(x)
        return x, trans_pts, trans_feat


class PointNetPartSegmentation(Module):
    def __init__(
        self,
        channels_in: int = 3,
        size_global_feature: int = 2048,
        use_pts_tnet: bool = True,
        use_feats_tnet: bool = True,
        align_normals: bool = False,
        num_classes: int = 16,
        num_classes_part: int = 50,
    ) -> None:
        """Create a new instance of PointNet Part Segmentation Network.

        Args:
            channels_in (optional): the size of first convolutional layer. Defaults to 3.
            size_global_feature (optional): the size for the global feature. Defaults to 2048.
            use_pts_tnet (optional): if True use the points T-Net. Defaults to True.
            use_feats_tnet (optional): if True use the features T-Net. Defaults to True.
            align_normals (optional): if True rotate normals using the T-Net. Defaults to True.
            num_classes (optional): the number of classes for the categorical vector. Defaults to 16.
            num_classes_part: (optional): the number of overal part segmentation classes Defaults to 50.
        """

        super(PointNetPartSegmentation, self).__init__()

        self.num_classes = num_classes
        self.num_classes_part = num_classes_part
        self.use_pts_tnet = use_pts_tnet
        self.use_feat_tnet = use_feats_tnet
        self.size_global_feature = size_global_feature
        self.align_normals = align_normals

        if self.use_pts_tnet:
            self.tnet_pts = TNet(channels_in=channels_in, dimension=3)

        if self.use_feat_tnet:
            self.tnet_feats = TNet(channels_in=128, dimension=128)

        self.conv_1 = Conv1d(channels_in, 64, 1)
        self.conv_2 = Conv1d(64, 128, 1)
        self.conv_3 = Conv1d(128, 128, 1)
        self.conv_4 = Conv1d(128, 512, 1)
        self.conv_5 = Conv1d(512, self.size_global_feature, 1)
        self.bn_1 = BatchNorm1d(64)  # type: ignore
        self.bn_2 = BatchNorm1d(128)  # type: ignore
        self.bn_3 = BatchNorm1d(128)  # type: ignore
        self.bn_4 = BatchNorm1d(512)  # type: ignore
        self.bn_5 = BatchNorm1d(self.size_global_feature)  # type: ignore

        size_in_conv = self.size_global_feature * 2 + self.num_classes + 64 + 128 * 2 + 512
        self.conv_seg_1 = Conv1d(size_in_conv, 256, 1)
        self.conv_seg_2 = Conv1d(256, 256, 1)
        self.conv_seg_3 = Conv1d(256, 128, 1)
        self.conv_seg_4 = Conv1d(128, num_classes_part, 1)
        self.bn_seg_1 = BatchNorm1d(256)  # type: ignore
        self.bn_seg_2 = BatchNorm1d(256)  # type: ignore
        self.bn_seg_3 = BatchNorm1d(128)  # type: ignore

    def forward(
        self, x: Tensor, vector_cat: Tensor
    ) -> Tuple[Tensor, Union[Tensor, None], Union[Tensor, None]]:
        """Process input point clouds and produces per points claffication scores.

        Args:
            x: The input clouds with shape (B, NUM_PTS, DIM).
            vector_cat: the one hot encoding categorical vector (B, NUM_CLASSES).

        Returns:
            A tuple containing:
                - The computed classification scores with shape (B, NUM_PTS, NUM_CLASSES_PART).
                - The learned transformation matrix for the input points.
                - The learned transformation matrix for the features.
        """
        num_points = x.shape[1]
        x = torch.transpose(x, 2, 1)

        if self.use_pts_tnet:
            trans_pts = self.tnet_pts(x)
            x = self.tnet_pts.transform(x, trans_pts, self.align_normals, False)
        else:
            trans_pts = None

        feat_1 = F.relu(self.bn_1(self.conv_1(x)))
        feat_2 = F.relu(self.bn_2(self.conv_2(feat_1)))
        feat_3 = F.relu(self.bn_3(self.conv_3(feat_2)))

        if self.use_feat_tnet:
            trans_feat = self.tnet_feats(feat_3)
            feat_3 = self.tnet_feats.transform(feat_3, trans_feat, False, True)
        else:
            trans_feat = None

        feat_4 = F.relu(self.bn_4(self.conv_4(feat_3)))
        feat_5 = self.bn_5(self.conv_5(feat_4))
        global_feature = torch.max(feat_5, 2, keepdim=True)[0]
        global_feature = global_feature.reshape(-1, self.size_global_feature)

        global_feature = torch.cat([global_feature, vector_cat], 1)
        global_feature = global_feature.unsqueeze(-1)
        global_feature = repeat(global_feature, "b ch n -> b ch (n r)", r=num_points)

        feats_concat = torch.cat([global_feature, feat_1, feat_2, feat_3, feat_4, feat_5], 1)
        feats_seg = F.relu(self.bn_seg_1(self.conv_seg_1(feats_concat)))
        feats_seg = F.relu(self.bn_seg_2(self.conv_seg_2(feats_seg)))
        feats_seg = F.relu(self.bn_seg_3(self.conv_seg_3(feats_seg)))
        feats_pts = self.conv_seg_4(feats_seg)
        feats_pts = torch.transpose(feats_pts, 2, 1)

        feats_pts = feats_pts.reshape(-1, num_points, self.num_classes_part)
        return feats_pts, trans_pts, trans_feat



from collections import defaultdict
from pathlib import Path
from statistics import mean
from typing import Dict, List, Tuple

import torch

from pycarus.geometry.pcd import read_pcd
from pycarus.metrics.chamfer_distance import chamfer
from pycarus.metrics.emd_distance import emd
from pycarus.metrics.f_score import f_score
from pycarus.utils import progress_bar

T_RESULTS_DICT = Dict[str, List[float]]


def _get_gts_and_preds_paths(gt_dir: Path, pred_dir: Path) -> Tuple[List[Path], List[Path]]:
    """Get ground-truths and predictions file path.

    Args:
        gt_dir: The path to the directory with all the ground-truths.
        pred_dir: The path to the directory with all the predictions.

    Returns:
        - A list with all ground-truth paths.
        - A list with all predictions paths.
    """
    gt_files = []
    pred_files = []
    categories_dirs = [subdir for subdir in gt_dir.iterdir() if subdir.is_dir()]
    for category_dir in categories_dirs:
        for gt_file in category_dir.glob("*.ply"):
            gt_files.append(gt_file)
            pred_files.append(pred_dir / category_dir.name / gt_file.name)

    return gt_files, pred_files


def _compute_single_results(
    gt_dir: Path,
    pred_dir: Path,
    results_dir: Path,
    batch_size: int,
    gpu: bool,
    compute_emd: bool,
) -> Tuple[T_RESULTS_DICT, T_RESULTS_DICT, T_RESULTS_DICT, T_RESULTS_DICT]:
    """Compute results for every single item and save them in the
    proper results file.

    Args:
        gt_dir: The path to the directory with all the ground-truths.
        pred_dir: The path to the directory with all the predictions.
        results_dir: The path to the output directory.
        batch_size: The batch size to use for computations.
        gpu: Whether to use or not gpu for computations.
        compute_emd: Whether to compute or not emd.

    Returns:
        - A dict with cd_p results for all categories.
        - A dict with cd_t results for all categories.
        - A dict with f1_score results for all categories.
        - A dict with emd results for all categories.
    """
    single_results_file = results_dir / "single_results.csv"
    if single_results_file.exists():
        single_results_file.unlink()

    with open(single_results_file, "wt") as f:
        f.write("CATEGORY/NAME,CD_P,CD_T,F1-SCORE")
        f.write(",EMD\n" if compute_emd else "\n")

    cd_ps = defaultdict(list)
    cd_ts = defaultdict(list)
    f1_scores = defaultdict(list)
    emds = defaultdict(list)

    gt_files, pred_files = _get_gts_and_preds_paths(gt_dir, pred_dir)

    num_batches = len(gt_files) // batch_size
    num_batches = num_batches + 1 if len(gt_files) % batch_size != 0 else num_batches

    for batch in progress_bar(range(num_batches)):
        start = batch * batch_size
        end = start + batch_size
        gt_files_batch = gt_files[start:end]
        pred_files_batch = pred_files[start:end]

        gts = torch.stack([read_pcd(gt_file) for gt_file in gt_files_batch], dim=0)
        preds = torch.stack([read_pcd(pred_file) for pred_file in pred_files_batch], dim=0)

        if gpu:
            gts = gts.cuda()
            preds = preds.cuda()

        _, cds_pred_gt, cds_gt_pred, _, _ = chamfer(preds, gts, squared=False)
        _, cds_pred_gt_sq, cds_gt_pred_sq, _, _ = chamfer(preds, gts, squared=True)
        f1s, _, _ = f_score(preds.cpu(), gts.cpu(), threshold=0.01)
        if compute_emd:
            emds_, _ = emd(preds, gts, eps=0.004, iterations=3000, squared=False, reduction="mean")

        for i, gt_file in enumerate(gt_files_batch):
            category = gt_file.parent.name
            shape_name = gt_file.stem

            cd_p = float(((cds_gt_pred[i] + cds_pred_gt[i]) / 2).item())
            cd_ps["all"].append(cd_p)
            cd_ps[category].append(cd_p)

            cd_t = float((cds_gt_pred_sq[i] + cds_pred_gt_sq[i]).item())
            cd_ts["all"].append(cd_t)
            cd_ts[category].append(cd_t)

            if compute_emd:
                emd_ = float(emds_[i].item())
                emds["all"].append(emd_)
                emds[category].append(emd_)

            f1_score = float(f1s[i].item())
            f1_scores["all"].append(f1_score)
            f1_scores[category].append(f1_score)

            with open(single_results_file, "at") as f:
                line = f"{category}/{shape_name},{cd_p:.8f},{cd_t:.8f},{f1_score:.8f}"
                line += f",{emd_:.8f}\n" if compute_emd else "\n"
                f.write(line)

    return cd_ps, cd_ts, f1_scores, emds


def _compute_global_results(
    results_dir: Path,
    cd_ps: T_RESULTS_DICT,
    cd_ts: T_RESULTS_DICT,
    f1_scores: T_RESULTS_DICT,
    emds: T_RESULTS_DICT,
    compute_emd: bool,
) -> None:
    """Compute global results and save them in the proper results file.

    Args:
        results_dir: The path to the output directory.
        cd_ps: The dict with cd_p results for all categories.
        cd_ts: The dict with cd_t results for all categories.
        f1_scores: The dict with f1_score results for all categories.
        emds: The dict with emd results for all categories.
        compute_emd: Whether to compute or not emd.
    """
    global_results_file = results_dir / "global_results.csv"
    if global_results_file.exists():
        global_results_file.unlink()

    with open(global_results_file, "wt") as f:
        f.write("CATEGORY,CD_P,CD_T,F1-SCORE")
        f.write(",EMD\n" if compute_emd else "\n")

    for category in sorted(cd_ps.keys()):
        if category != "all":
            m_cd_p = mean(cd_ps[category])
            m_cd_t = mean(cd_ts[category])
            m_f1_score = mean(f1_scores[category])
            if compute_emd:
                m_emd_ = mean(emds[category])

            with open(global_results_file, "at") as f:
                f.write(f"{category},{m_cd_p:.8f},{m_cd_t:.8f},{m_f1_score:.8f}")
                f.write(f",{m_emd_:.8f}\n" if compute_emd else "\n")

    with open(global_results_file, "at") as f:
        f.write("\n\n")
        f.write(f"MEAN CD_P,{mean(cd_ps['all']):.8f}\n")
        f.write(f"MEAN CD_T,{mean(cd_ts['all']):.8f}\n")
        f.write(f"MEAN F1 SCORE,{mean(f1_scores['all']):.8f}\n")
        if compute_emd:
            f.write(f"MEAN EMD,{mean(emds['all']):.8f}\n")


def pcd_reconstruction_evaluation(
    gt_dir: Path,
    pred_dir: Path,
    results_dir: Path,
    batch_size: int = 512,
    gpu: bool = True,
    compute_emd: bool = True,
) -> None:
    """Evaluate point cloud reconstruction results.

    The function requires the ground-truth clouds and the predicted ones to be
    organized in two separate directories with the same structure. Each directory
    must contain one sub-directory for each category and each subdirectory must
    contain one .ply file for each cloud. The function expects the name of the files
    in the "gt_dir" and "pred_dir" directories to be matching in order to perform
    the comparison.

    Four metrics are computed: the two Chamfer Distances proposed in the paper
    "Variational Relational Point Completion Network.", Earth Mover's Distance and F1 score.
    Results are saved into two files ("global_results.csv" and "single_results.csv")
    inside the given "result_dir" directory.

    Args:
        gt_dir: Directory with ground-truth clouds.
        pred_dir: Directory with predicted clouds.
        results_dir: Directory where the .csv files with the results will be saved.
        batch_size: The batch size to use for computing metrics. Defaults to 512.
        gpu: Whether to use gpu or not to compute metrics. Defaults to True.
        compute_emd: Whether to compute or not EMD. Defaults to True.

    Raises:
        ValueError: If one prediction is missing.
    """
    results_dir.mkdir(parents=True, exist_ok=True)

    cd_ps, cd_ts, f1_scores, emds = _compute_single_results(
        gt_dir,
        pred_dir,
        results_dir,
        batch_size,
        gpu,
        compute_emd,
    )

    _compute_global_results(results_dir, cd_ps, cd_ts, f1_scores, emds, compute_emd)



import math
from typing import Tuple

import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.nn.parallel
import torch.utils.data
from torch import Tensor


class PCN_encoder(nn.Module):
    """Model class respresenting the Point Completion Network Encoder as proposed in:

    Yuan, W., Khot, T., Held, D., Mertz, C., & Hebert, M. (2018).
    Pcn: Point completion network.
    In 2018 International Conference on 3D Vision (3DV) (pp. 728-737). IEEE.
    """

    def __init__(self, size_latent: int = 1024) -> None:
        """Create an instance of Point Completion Network Encoder.

        Args:
            size_latent: the dimension for the latent code.
        """
        super(PCN_encoder, self).__init__()
        self.conv1 = nn.Conv1d(3, 128, 1)
        self.conv2 = nn.Conv1d(128, 256, 1)
        self.conv3 = nn.Conv1d(512, 512, 1)
        self.conv4 = nn.Conv1d(512, size_latent, 1)

    def forward(self, x: Tensor) -> Tensor:
        """Process input point cloud and produce the latent codes.

        Args:
            x: The input point cloud with shape (BATCH, NUM_POINTS, 3). The method internally
            transposes the data tensor to correctly operate with PyTorch convetion (B CH N).

        Returns:
            The latent codes with shape (BATCH, SIZE_LATENT).
        """
        x = x.transpose(2, 1)
        batch_size, _, num_points = x.size()
        x = F.relu(self.conv1(x))
        x = self.conv2(x)
        global_feature, _ = torch.max(x, 2)
        x = torch.cat(
            (x, global_feature.reshape(batch_size, -1, 1).repeat(1, 1, num_points).contiguous()), 1
        )
        x = F.relu(self.conv3(x))
        x = self.conv4(x)
        global_feature, _ = torch.max(x, 2)
        return global_feature.reshape(batch_size, -1)


class PCN_decoder(nn.Module):
    """Model class respresenting the Point Completion Network Decoder as proposed in:

    Yuan, W., Khot, T., Held, D., Mertz, C., & Hebert, M. (2018).
    Pcn: Point completion network.
    In 2018 International Conference on 3D Vision (3DV) (pp. 728-737). IEEE.
    """

    def __init__(
        self,
        num_pcd_coarse: int,
        num_pcd_fine: int,
        scale: float,
        num_cat_feature: int,
        device: torch.device,
    ) -> None:
        """Create an instance of Point Completion Network Decoder.

        Args:
            num_pcd_coarse: the number of points for the coarse point cloud.
            num_pcd_fine: the number of points for the fine point cloud.
            scale: the scale for the folded grid.
            num_cat_feature: the size of the input embedding.
            device: the device to use for the grid.
        """
        super(PCN_decoder, self).__init__()

        self.num_coarse = num_pcd_coarse
        self.num_fine = num_pcd_fine
        self.fc1 = nn.Linear(1024, 1024)
        self.fc2 = nn.Linear(1024, 1024)
        self.fc3 = nn.Linear(1024, self.num_coarse * 3)

        self.scale = scale
        grid = self.__gen_grid_up(2 ** (int(math.log2(scale))), 0.05)
        self.device = device
        self.grid = grid.to(self.device).contiguous()
        self.conv1 = nn.Conv1d(num_cat_feature, 512, 1)
        self.conv2 = nn.Conv1d(512, 512, 1)
        self.conv3 = nn.Conv1d(512, 3, 1)

    @staticmethod
    def __gen_grid_up(up_ratio: int, grid_size: float) -> Tensor:
        """Generate coordinates grid.

        Args:
            up_ratio: The scale of the grid.
            grid_size: The coordinate reference frame [-grid_size, grid_size].

        Returns:
            The generated grid.
        """
        sqrted = int(math.sqrt(up_ratio)) + 1
        num_x = 1
        num_y = 1
        for i in range(1, sqrted + 1).__reversed__():
            if (up_ratio % i) == 0:
                num_x = i
                num_y = int(up_ratio // i)
                break

        grid_x = torch.linspace(-grid_size, grid_size, steps=num_x)
        grid_y = torch.linspace(-grid_size, grid_size, steps=num_y)

        x, y = torch.meshgrid(grid_x, grid_y, indexing="ij")  # x, y shape: (2, 1)
        grid = torch.stack([x, y], dim=-1).reshape(-1, 2).transpose(0, 1).contiguous()
        return grid

    def forward(self, x: Tensor) -> Tuple[Tensor, Tensor]:
        """Generate the coarse and fine point clouds from the embedding.

        Args:
            x: The input latent codes with shape (BATCH, SIZE_LATENT).

        Returns:
            A Tuple containing:
            - The generated coarse point cloud with shape (BATCH, 3, NUM_PCD_COARSE).
            - The generated fine point cloud with shape (BATCH, 3, NUM_PCD_FINE).
        """
        batch_size = x.size()[0]
        coarse = F.relu(self.fc1(x))
        coarse = F.relu(self.fc2(coarse))
        coarse = self.fc3(coarse).reshape(-1, 3, self.num_coarse)

        grid = self.grid.clone().detach()
        grid_feat = grid.unsqueeze(0).repeat(batch_size, 1, self.num_coarse).contiguous()
        grid_feat = grid_feat.to(self.device)

        point_feat = (
            (
                (coarse.transpose(1, 2).contiguous())  # type: ignore
                .unsqueeze(2)
                .repeat(1, 1, self.scale, 1)
                .reshape(-1, self.num_fine, 3)
            )
            .transpose(1, 2)
            .contiguous()
        )

        global_feat = x.unsqueeze(2).repeat(1, 1, self.num_fine)

        feat = torch.cat((grid_feat, point_feat, global_feat), 1)

        center = (
            (
                (coarse.transpose(1, 2).contiguous())  # type: ignore
                .unsqueeze(2)
                .repeat(1, 1, self.scale, 1)
                .reshape(-1, self.num_fine, 3)
            )
            .transpose(1, 2)
            .contiguous()
        )

        fine = self.conv3(F.relu(self.conv2(F.relu(self.conv1(feat))))) + center
        return coarse, fine


class PCNNetwork(nn.Module):
    """Model class respresenting the Point Completion Network as proposed in:

    Yuan, W., Khot, T., Held, D., Mertz, C., & Hebert, M. (2018).
    Pcn: Point completion network.
    In 2018 International Conference on 3D Vision (3DV) (pp. 728-737). IEEE.
    """

    def __init__(
        self, device: torch.device, num_points: int = 2048, num_points_coarse: int = 1024
    ) -> None:
        """Create an instance of Point Completion Network.

        Args:
            device: the device to use for the grid to fold.
            num_points: The number of points for the fine point cloud. Defaults to 2048.
            num_points_coarse: The number of points for the coarse point cloud. Defaults to 1024.
        """
        super(PCNNetwork, self).__init__()
        self.num_coarse = num_points_coarse
        self.num_points = num_points
        self.scale = self.num_points // num_points_coarse
        self.cat_feature_num = 2 + 3 + 1024
        self.device = device

        self.encoder = PCN_encoder()
        self.decoder = PCN_decoder(
            self.num_coarse, self.num_points, self.scale, self.cat_feature_num, self.device
        )

    def forward(self, x: Tensor) -> Tuple[Tensor, Tensor]:
        """Perform point cloud completion on the input point cloud.

        Args:
            x: The input point cloud with shape (BATCH, NUM_POINTS, 3). The method internally
            transposes the data tensor to correctly operate with PyTorch convetion (B CH N).

        Returns:
            A Tuple containing:
            - The generated coarse point cloud with shape (BATCH, NUM_PCD_COARSE, 3).
            - The generated fine point cloud with shape (BATCH, NUM_PCD_FINE, 3).

            The method internally transposes the output data tensors.
        """
        feat = self.encoder(x)
        coarse, fine = self.decoder(feat)
        coarse = coarse.transpose(1, 2).contiguous()
        fine = fine.transpose(1, 2).contiguous()

        return coarse, fine

    def __repr__(self) -> str:
        layer_str = ""
        for name, param in self.named_parameters():
            if "kernel" in name:
                layer_str += "Name: {} - Shape {}".format(name, param.transpose(2, 1).shape) + "\n"

        return super(PCNNetwork, self).__repr__() + layer_str  # type: ignore



from typing import List, Tuple

import torch
import torch.nn as nn
import torch.nn.functional as F
from einops import repeat  # type: ignore
from pytorch3d.ops import ball_query, knn_gather, knn_points  # type: ignore
from torch import Tensor
from torch.nn import BatchNorm2d, Conv2d, Module, ModuleList

from pycarus.geometry.pcd import farthest_point_sampling


def replace_minus_one(x: Tensor) -> Tensor:
    """Replace -1 pad with the index of the first element of the tensor.

    Args:
        x: indices tensor returned from ball_query with shape (B, NUM_PTS, K).

    Returns:
        The padded indices with new values with shape (B, NUM_PTS, K).
    """
    pad_values = x[:, :, 0]
    pad_values = repeat(pad_values, "b n -> b n r", r=x.shape[-1])

    indices_pad = x == -1
    x[indices_pad] = pad_values[indices_pad]

    return x


class PointNetSetAbstraction(Module):
    def __init__(
        self,
        num_point: int,
        radius: float,
        num_samples: int,
        in_channel: int,
        mlp: List[int],
        group_all: bool,
    ) -> None:
        super(PointNetSetAbstraction, self).__init__()
        self.num_point = num_point
        self.radius = radius
        self.num_sample = num_samples
        self.mlp_convs = ModuleList()
        self.mlp_bns = ModuleList()
        last_channel = in_channel
        for out_channel in mlp:
            self.mlp_convs.append(Conv2d(last_channel, out_channel, 1))
            self.mlp_bns.append(BatchNorm2d(out_channel))  # type: ignore
            last_channel = out_channel
        self.group_all = group_all

    def forward(self, xyz: Tensor, features: Tensor) -> Tuple[Tensor, Tensor]:
        """Process input point cloud and produce point net++ features.

        Args:
            xyz: the input points coordinates with shape (B, NUM_PTS, 3).
            features: the input features data with shaoe (B, NUM_PTS, D).

        Returns:
            A tuple containing:
             - The sampled points coordinates with fps with shape (B, NUM_POINTS, 3).
             - The concatenated points feature data (B, NUM_POINTS, D').
        """
        if self.group_all:
            size_b, num_pts, ch = xyz.shape
            grouped_xyz = xyz.reshape(size_b, 1, num_pts, ch)

            if features is not None:
                features_grouped = torch.cat(
                    [grouped_xyz, features.reshape(size_b, 1, num_pts, -1)],
                    dim=-1,
                )
            else:
                features_grouped = grouped_xyz

            xyz_grouped = torch.zeros(size_b, 1, ch).to(xyz.device)
        else:
            raise NotImplementedError("Function not yet implemented.")

        # new_xyz: sampled points position data, [B, npoint, C]
        # new_points: sampled points data, [B, npoint, nsample, C+D]
        features_grouped = features_grouped.permute(0, 3, 2, 1)  # [B, C+D, nsample,npoint]
        for i, conv in enumerate(self.mlp_convs):
            bn = self.mlp_bns[i]
            features_grouped = F.relu(bn(conv(features_grouped)))

        features_out = torch.max(features_grouped, 2)[0]
        features_out = features_out.transpose(2, 1)

        return xyz_grouped, features_out


class PointNetSetAbstractionMsg(Module):
    def __init__(
        self,
        num_point: int,
        list_radius: List[float],
        list_num_samples: List[int],
        in_channel: int,
        mlp_list: List[List[int]],
    ) -> None:
        """Create a new instance of PointNet++ Set Abstraction Layer MSG.

        Args:
            num_point: the numbe of points to sample with fps.
            list_radius: the list of radius to use for ball queries.
            list_num_samples: the list of max number of samples to keep for ball queries.
            in_channel: the number of input channels.
            mlp_list: the list of MLP modules channels.
        """
        super(PointNetSetAbstractionMsg, self).__init__()

        self.num_point = num_point
        self.radius_list = list_radius
        self.nsample_list = list_num_samples
        self.conv_blocks = ModuleList()
        self.bn_blocks = ModuleList()

        for i in range(len(mlp_list)):
            convs = ModuleList()
            bns = ModuleList()
            last_channel = in_channel + 3
            for out_channel in mlp_list[i]:
                convs.append(Conv2d(last_channel, out_channel, 1))
                bns.append(BatchNorm2d(out_channel))  # type: ignore
                last_channel = out_channel
            self.conv_blocks.append(convs)
            self.bn_blocks.append(bns)

    def forward(self, xyz: Tensor, features: Tensor) -> Tuple[Tensor, Tensor]:
        """Process input point cloud and produce point net++ features.

        Args:
            xyz: the input points coordinates with shape (B, NUM_PTS, 3).
            features: the input features data with shaoe (B, NUM_PTS, D).

        Returns:
            A tuple containing:
             - The sampled points coordinates with fps with shape (B, NUM_POINTS, 3).
             - The concatenated points feature data (B, NUM_POINTS, D').
        """
        _, _, C = xyz.shape

        xyz_ss = farthest_point_sampling(xyz, self.num_point)
        list_features_out = []

        for i, radius in enumerate(self.radius_list):
            _, indices_group, _ = ball_query(xyz_ss, xyz, K=self.nsample_list[i], radius=radius)
            indices_group = replace_minus_one(indices_group)
            xyz_grouped = knn_gather(xyz, indices_group)
            xyz_grouped -= xyz_ss.reshape(-1, self.num_point, 1, C)

            if features is not None:
                features_grouped = knn_gather(features, indices_group)
                features_grouped = torch.cat([features_grouped, xyz_grouped], dim=-1)
            else:
                features_grouped = xyz_grouped

            features_grouped = features_grouped.permute(0, 3, 2, 1)  # [B, D, K, S]
            for j in range(len(self.conv_blocks[i])):
                conv = self.conv_blocks[i][j]
                bn = self.bn_blocks[i][j]
                features_grouped = F.relu(bn(conv(features_grouped)))

            list_features_out.append(torch.max(features_grouped, 2)[0])  # [B, D', S]

        features_out = torch.cat(list_features_out, dim=1).transpose(2, 1)
        return xyz_ss, features_out


class PointNetFeaturePropagation(Module):
    def __init__(self, in_channel: int, mlp: List[int]) -> None:
        """Create a new instance of PointNetFeaturePropagation Layer.

        Args:
            in_channel: the size of first input channels.
            mlp: the size of output channels for each mlp.
        """
        super(PointNetFeaturePropagation, self).__init__()
        self.mlp_convs = ModuleList()
        self.mlp_bns = ModuleList()
        last_channel = in_channel

        for out_channel in mlp:
            self.mlp_convs.append(nn.Conv1d(last_channel, out_channel, 1))
            self.mlp_bns.append(nn.BatchNorm1d(out_channel))  # type: ignore
            last_channel = out_channel

    def forward(self, xyz_s: Tensor, xyz_t: Tensor, feats_s: Tensor, feats_t: Tensor) -> Tensor:
        """Process input point clouds and produces per interpolated features.

        Input:
            xyz_s: the input points coordinates with shape (B, NUM_PTS_S, 3).
            xyz_t: the sampled input points position with shape (B, NUM_PTS_T, 3).
            feats_s: xyz_s per point features with shape (B, NUM_PTS_S, D).
            feats_t: xyz_t per point features with shape (B, NUM_PTS_T, D).

        Return:
            The interpolead features with shape (B, NUM_PTS_S, D').
        """

        _, N, _ = xyz_s.shape
        _, S, _ = xyz_t.shape

        if S == 1:
            feats_inter = feats_t.repeat(1, N, 1)
        else:
            dists, indices, _ = knn_points(xyz_s, xyz_t, K=3)

            dist_recip = 1.0 / (dists + 1e-8)
            norm = torch.sum(dist_recip, dim=2, keepdim=True)
            weight = dist_recip / norm
            feats_inter = torch.sum(
                knn_gather(feats_t, indices) * weight.reshape(-1, N, 3, 1), dim=2
            )

        if feats_s is not None:
            features = torch.cat([feats_s, feats_inter], dim=-1)
        else:
            features = feats_inter

        features = features.transpose(2, 1)
        for i, conv in enumerate(self.mlp_convs):
            bn = self.mlp_bns[i]
            features = F.relu(bn(conv(features)))

        features_out = features.transpose(2, 1)
        return features_out



from pathlib import Path
from typing import Callable, List, Tuple, Union

import torch
from torch import Tensor
from torch.utils.data import Dataset

from pycarus.geometry.pcd import read_pcd
from pycarus.transforms.var import Compose


class PlyDataset(Dataset):
    def __init__(
        self,
        root: Union[str, Path],
        split: str,
        transforms: List[Callable] = [],
    ) -> None:
        """Generic dataset to load point clouds stored as PLY files.

        Point clouds are expected to be stored in separate subfolders inside "root",
        one subfolder for each class. Each subfolder is expected to contain
        separate subfolders for each split.
        Class IDs are computed by sorting the class names in alphabetical order.

        Args:
            root: The path to the directory with the dataset.
            transforms: The list of transforms to be applied to the point clouds.
        """
        super().__init__()

        self.root = Path(root)

        self.classes = sorted([s.name for s in self.root.iterdir()])

        self.pcds: List[Tensor] = []
        self.labels: List[Tensor] = []

        for i, c in enumerate(self.classes):
            class_path = self.root / c / split

            items_paths = sorted(list(class_path.glob("*.ply")))
            for p in items_paths:
                self.pcds.append(read_pcd(p))
                self.labels.append(torch.tensor(i, dtype=torch.long))

        self.transform = Compose(transforms)

    def __len__(self) -> int:
        """Return the number of pcds in the dataset.

        Returns:
            The number of pcds in the dataset.
        """
        return len(self.pcds)

    def __getitem__(self, index: int) -> Tuple[Tensor, Tensor]:
        """Return the pcd and the label at the given index.

        Args:
            index: The index of the required element.

        Returns:
            - A tensor with the point cloud with shape (NUM_POINTS, 3).
            - A tensor with the label.
        """
        pcd = self.transform(self.pcds[index])
        return pcd, self.labels[index]

    def get_class(self, label: int) -> str:
        """Return the name of the class with the given label.

        Args:
            label: The numeric label of the required class.

        Returns:
            The name of the class corresponding to the given label.
        """
        return self.classes[label]



from setuptools import setup
from torch.utils.cpp_extension import BuildExtension, CUDAExtension

setup(
    name="chamfercuda",
    ext_modules=[
        CUDAExtension(  # type: ignore
            "chamfercuda",
            [
                "/".join(__file__.split("/")[:-1] + ["chamfer_cuda_cpp.cpp"]),
                "/".join(__file__.split("/")[:-1] + ["chamfer_cuda_cu.cu"]),
            ],
        ),
    ],
    cmdclass={"build_ext": BuildExtension},
)



from pathlib import Path
from typing import Callable, List, Tuple, cast

import torch
from torch import Tensor
from torch.utils.data import Dataset

from pycarus.transforms.var import Compose

T_ITEM = Tuple[str, str, Tensor, Tensor]


class Pcn(Dataset):
    """Class implementing the Point Completion Network dataset as proposed in:

    Yuan, W., Khot, T., Held, D., Mertz, C., & Hebert, M. (2018, September).
    Pcn: Point completion network.
    In 2018 International Conference on 3D Vision (3DV) (pp. 728-737). IEEE.
    """

    def __init__(
        self,
        root: Path,
        split: str,
        use_original_incomplete: bool = False,
        use_original_complete: bool = True,
        transforms_incomplete: List[Callable] = [],
        transforms_complete: List[Callable] = [],
        transforms_all: List[Callable] = [],
    ) -> None:
        """Create an instance of PCN dataset.

        Args:
            root: The path to the root directory of the dataset.
            split: The split to use (only "train", "val" and "test" allowed).
            use_original_incomplete: If True, use original incomplete clouds, which have
                a different number of points for each item. Otherwise, use incomplete clouds
                sampled with FPS to obtain 2048 points. Defaults to False.
            use_original_complete: If True, use original complete clouds, which have
                16384 points for each item. Otherwise, use complete clouds sampled with
                FPS to obtain 2048 points. Defaults to True.
            transforms_incomplete: The list of transforms to apply to incomplete pcds. Defaults to [].
            transforms_complete: The list of transforms to apply to complete pcds. Defaults to [].
            transforms_all: The list of transforms to apply to all pcds. Defaults to [].
        """
        super().__init__()

        assert split in ["train", "val", "test"]

        self.root = root / split
        self.num_items = len(list(self.root.glob("*.pt")))

        self.use_original_incomplete = use_original_incomplete
        self.use_original_complete = use_original_complete

        self.transform_complete = Compose(transforms_complete)
        self.transform_incomplete = Compose(transforms_incomplete)
        self.transform_all = Compose(transforms_all)

    def __len__(self) -> int:
        """Return the number of items in the dataset.

        Returns:
            The number of items in the dataset.
        """
        return self.num_items

    def __getitem__(self, index: int) -> T_ITEM:
        """Return the item at the given index.

        The number of points of incomplete and complete clouds depends on the
        flags "use_original_incomplete" and "use_original_complete" set when
        creating the dataset.

        Args:
            index: The index of the required item.

        Returns:
            - The category of the item.
            - The name of the item.
            - A tensor with the incomplete point cloud with shape (NUM_POINTS, 3).
            - A tensor with the complete point cloud with shape (NUM_POINTS, 3).
        """
        item_path = self.root / f"{index}.pt"
        item = torch.load(item_path)  # type: ignore

        category = cast(str, item["category"])
        name = cast(str, item["name"])

        if self.use_original_incomplete:
            incomplete = cast(Tensor, item["original_incomplete"])
        else:
            incomplete = cast(Tensor, item["incomplete_fps"])

        if self.use_original_complete:
            complete = cast(Tensor, item["original_complete"])
        else:
            complete = cast(Tensor, item["complete_fps"])

        incomplete = self.transform_all(self.transform_incomplete(incomplete))
        complete = self.transform_all(self.transform_complete(complete))

        return category, name, incomplete, complete



from pathlib import Path
from typing import Callable, List, Tuple, Union

import torch
from torch import Tensor
from torch.utils.data import Dataset

from pycarus.geometry.mesh import read_mesh
from pycarus.transforms.var import Compose


class Manifold40(Dataset):
    """Manifold40 Dataset, as proposed in
    Hu, Shi-Min, et al. "Subdivision-Based Mesh Convolution Networks."
    """

    def __init__(
        self,
        root: Union[Path, str],
        split: str,
        only_version: int = -1,
        vertices_transforms: List[Callable] = [],
        triangles_transforms: List[Callable] = [],
        pad_vertices_to: int = 6200,
        pad_triangles_to: int = 12_300,
    ) -> None:
        """Create an instance of Manifold40 dataset.

        Args:
            root: The path to the root directory of the dataset.
            split: The split to use (only "train", "val" and "test" allowed).
            only_version: The original dataset contains 10 versions of each shape,
                numbered from 0 to 9. This argument can be used to get only the
                specified version. If it is equal to -1, then all versions are returned.
                All versions are always returned for the test split.
            vertices_transforms: The transform to apply to the vertices of
                each item. Defaults to [].
            triangles_transforms: The transform to apply to the triangles of
                each item. Defaults to [].
            pad_vertices_to: If greater than 0, the vertices will be padded with
                zeros to obtain a tensor with the required number of vertices.
                Defaults to 6200, which is the maximum number of vertices
                of a shape in the dataset.
            pad_triangles_to: If greater than 0, the triangles will be padded with
                zeros to obtain a tensor with the required number of triangles.
                Defaults to 12_300, which is the maximum number of triangles
                of a shape in the dataset.
        """
        if split not in ["train", "val", "test"]:
            raise ValueError("Only train, val and test splits are available.")

        self.root = Path(root)

        self.classes = sorted([s.name for s in self.root.iterdir()])

        self.paths: List[Path] = []
        self.labels: List[Tensor] = []

        for i, c in enumerate(self.classes):
            class_path = self.root / c / split

            items_paths = sorted(list(class_path.glob("*.obj")))
            for p in items_paths:
                if "test" in split or only_version == -1 or f"-{only_version}.obj" in p.name:
                    self.paths.append(p)
                    self.labels.append(torch.tensor(i, dtype=torch.long))

        self.vertices_transform = Compose(vertices_transforms)
        self.triangles_transform = Compose(triangles_transforms)

        self.pad_vertices_to = pad_vertices_to
        self.pad_triangles_to = pad_triangles_to

    def __len__(self) -> int:
        """Return the number of meshes in the dataset.

        Returns:
            The number of meshes in the dataset.
        """
        return len(self.paths)

    def __getitem__(self, index: int) -> Tuple[Tensor, Tensor, Tensor, int, int]:
        """Return the vertices, the triangles and the label at the given index.

        Args:
            index: The index of the required element.

        Returns:
            - A tensor with the vertices with shape (N, D). D can be 3, 6 or 9
                depending on the availability of normals and colors.
            - A tensor with the triangles with shape (M, D). D can be 3 or 6
                depending on the availability of normals.
            - A tensor with the label.
            - The original number of vertices before padding (if performed).
            - The original number of triangles before padding (if performed).
        """
        vertices, triangles = read_mesh(self.paths[index])
        vertices = self.vertices_transform(vertices)
        triangles = self.triangles_transform(triangles)

        num_vertices = vertices.shape[0]
        num_triangles = triangles.shape[0]

        if self.pad_vertices_to > 0:
            padded_vertices = torch.zeros((self.pad_vertices_to, 3))
            padded_vertices[:num_vertices] = vertices
            vertices = padded_vertices

        if self.pad_triangles_to > 0:
            padded_triangles = torch.zeros((self.pad_triangles_to, 3))
            padded_triangles[:num_triangles] = triangles
            triangles = padded_triangles

        return vertices, triangles, self.labels[index], num_vertices, num_triangles

    def get_class(self, label: int) -> str:
        """Return the name of the class with the given label.

        Args:
            label: The numeric label of the required class.

        Returns:
            The name of the class corresponding to the given label.
        """
        return self.classes[label]



from typing import Tuple

import torch
from pytorch3d.ops.knn import knn_points  # type: ignore
from torch import Tensor, mean, sqrt

from pycarus.geometry.pcd import batchify, unbatchify


def chamfer(
    predictions: Tensor,
    groundtruth: Tensor,
    squared: bool = True,
) -> Tuple[Tensor, Tensor, Tensor, Tensor, Tensor]:
    """Compute the Chamfer Distance defined as in:

       Fan, H., Su, H., & Guibas, L. J. (2017).
       A point set generation network for 3d object reconstruction from a single image.
       In Proceedings of the IEEE conference on computer vision and pattern recognition (pp. 605-613).

    Args:
        prediction: The source point cloud(s) with shape ([B,] NUM_PTS_PRED, 3).
        groundtruth: The target point cloud(s) with shape ([B,] NUM_PTS_GT, 3).
        squared: If true return the squared euclidean distance.

    Returns:
        A tuple containing:
            - the chamfer distance with shape (B, ).
            - the accuracy, eg. point-wise distance from the prediction to the groundtruth with shape ([B,] NUM_PTS_PRED).
            - the completeness, eg. the distance from the groundtruth to the prediction with shape ([B,] NUM_PTS_GT).
            - the nearest neighbor indices from the prediction to the groundtruth with shape ([B,] NUM_PTS_PRED).
            - the nearest neighbor indices from the groundtruth to the prediction with shape ([B,] NUM_PTS_GT).
    """
    batched, [predictions, groundtruth] = batchify([predictions, groundtruth], 3)

    accuracy, indices_pred_gt, _ = knn_points(predictions, groundtruth)
    accuracy = accuracy.squeeze(-1)
    indices_pred_gt = indices_pred_gt.squeeze(-1)

    completeness, indices_gt_pred, _ = knn_points(groundtruth, predictions)
    completeness = completeness.squeeze(-1)
    indices_gt_pred = indices_gt_pred.squeeze(-1)

    if not squared:
        accuracy = sqrt(accuracy)
        completeness = sqrt(completeness)

    avg_accuracy = mean(accuracy, dim=-1)
    avg_completeness = mean(completeness, dim=-1)

    chamfer = avg_accuracy + avg_completeness

    if batched:
        chamfer, accuracy, completeness, indices_pred_gt, indices_gt_pred = unbatchify(
            [chamfer, accuracy, completeness, indices_pred_gt, indices_gt_pred]
        )

    return chamfer, accuracy, completeness, indices_pred_gt, indices_gt_pred


def chamfer_p(prediction: Tensor, groundtruth: Tensor) -> Tensor:
    """Compute the chamfer distance "P".

    The result is computed as:
        - _, cd_pred2gt, cd_gt2pred, _, _ = chamfer(prediction, groundtruth, squared=False)
        - result = (cd_pred2gt.mean(-1) + cd_gt2pred.mean(-1)) / 2.

    Args:
        prediction: The source point cloud(s) with shape ([B,] NUM_POINTS, 3).
        groundtruth: The target point cloud(s) with shape ([B,] NUM_POINTS, 3).

    Returns:
        The computed chamfer distance "P".
    """
    _, cds_pred_gt, cds_gt_pred, _, _ = chamfer(prediction, groundtruth, squared=False)
    cds_pred_gt = torch.mean(cds_pred_gt, dim=-1)
    cds_gt_pred = torch.mean(cds_gt_pred, dim=-1)

    return (cds_gt_pred + cds_pred_gt) / 2


def chamfer_t(prediction: Tensor, groundtruth: Tensor) -> Tensor:
    """Compute the chamfer distance "T".

    The result is computed as:
        - _, cd_pred2gt, cd_gt2pred, _, _ = chamfer(prediction, groundtruth, squared=True)
        - result = cd_pred2gt.mean(-1) + cd_gt2pred.mean(-1).

    Args:
        prediction: The source point cloud(s) with shape ([B,] NUM_POINTS, 3).
        groundtruth: The target point cloud(s) with shape ([B,] NUM_POINTS, 3).

    Returns:
        The computed chamfer distance "T".
    """
    _, cds_pred_gt_sq, cds_gt_pred_sq, _, _ = chamfer(prediction, groundtruth, squared=True)
    cds_pred_gt_sq = torch.mean(cds_pred_gt_sq, dim=-1)
    cds_gt_pred_sq = torch.mean(cds_gt_pred_sq, dim=-1)

    return cds_pred_gt_sq + cds_gt_pred_sq



from pathlib import Path
from typing import Dict, List, Tuple

import numpy as np
import torch

from pycarus.metrics.partseg_iou import PartSegmentationIoU


def _get_ious_from_file(
    pred_file: Path,
    use_only_category_logits: bool,
    category_to_parts_map: Dict[str, List[int]],
) -> Tuple[float, float]:
    """Compute instance IoU and class IoU from a prediction file.

    Args:
        pred_file: The path to the prediction file.
        use_only_category_logits: Whether to compute predictions considering only
                                  the logits corresponding to the parts belonging
                                  to the current class.
        category_to_parts_map: A dictionary that maps classes to parts labels.
                               E.G. {"airplane": [0, 1, 2, ...], ...}

    Returns:
        - Mean Class IoU: the average mIoU for each class.
        - Mean Instance IoU: the average IoU for each sample.
    """
    data = np.load(pred_file)
    labels = torch.tensor(data["labels"])
    logits = torch.tensor(data["logits"])

    metric = PartSegmentationIoU(use_only_category_logits, category_to_parts_map)

    for i in range(labels.shape[0]):
        lbl = labels[i].unsqueeze(0)
        lgt = logits[i].unsqueeze(0)
        metric(lgt, lbl)

    mIoUs, mean_class_IoU, mean_instance_IoU = metric.compute()

    return mean_class_IoU, mean_instance_IoU


def pcd_part_segmentation_evaluation(
    pred_files: List[Path],
    outdir: Path,
    use_only_category_logits: bool,
    category_to_parts_map: Dict[str, List[int]],
) -> None:
    """Evaluation for point cloud part segmentation.

    Each prediction file is expected to be produced by a different run
    and should be a .npz file containing one array labelled as "labels"
    and one array labelled as "logits". The "labels" array is a 2-D
    array with integer labels with shape (NUM_PCDS, NUM_POINTS_PER_PCD),
    while the "logits" array is a 3-D array with logits predicted by some
    model with shape (NUM_PCDS, NUM_POINTS_PER_PCD, NUM_CLASSES).

    Args:
        pred_files: The list of files with all the predictions.
        outdir: The path to the directory where results will be saved.
        use_only_category_logits: Whether to compute predictions considering only
                                  the logits corresponding to the parts belonging
                                  to the current category.
        category_to_parts_map: A dictionary that maps classes to parts labels.
                               E.G. {"airplane": [0, 1, 2, ...], ...}
    """
    class_ious: List[float] = []
    inst_ious: List[float] = []

    for pred_file in pred_files:
        class_iou, inst_iou = _get_ious_from_file(
            pred_file,
            use_only_category_logits,
            category_to_parts_map,
        )
        class_ious.append(class_iou)
        inst_ious.append(inst_iou)

    outdir.mkdir(parents=True, exist_ok=True)
    with open(outdir / "results.csv", "wt") as f:
        f.write("METRIC,MEAN,STD\n")

        if len(class_ious) > 1:
            class_iou_mean = np.mean(np.array(class_ious)) * 100
            class_iou_std = np.std(np.array(class_ious)) * 100
            f.write(f"CLASS mIoU,{class_iou_mean:.3f},{class_iou_std:.3f}\n")
            inst_iou_mean = np.mean(np.array(inst_ious)) * 100
            inst_iou_std = np.std(np.array(inst_ious)) * 100
            f.write(f"INSTANCE mIoU,{inst_iou_mean:.3f},{inst_iou_std:.3f}")
        else:
            f.write(f"CLASS mIoU,{class_ious[0] * 100:.3f},-\n")
            f.write(f"INSTANCE mIoU,{inst_ious[0] * 100:.3f},-")



from typing import Tuple, Union

import torch
from pytorch3d.ops.knn import knn_points  # type: ignore
from torch import Tensor, sqrt, sum, zeros

from pycarus.geometry.pcd import batchify, unbatchify


def f_score(
    predictions: Union[Tensor, None],
    groundtruth: Union[Tensor, None],
    dist_pred_gt: Union[Tensor, None] = None,
    dist_gt_pred: Union[Tensor, None] = None,
    threshold: float = 0.01,
) -> Tuple[Tensor, Tensor, Tensor]:
    """Compute the F1-Score using the treshold as defined in:

    Knapitsch, A., Park, J., Zhou, Q. Y., & Koltun, V. (2017).
    Tanks and temples: Benchmarking large-scale scene reconstruction.
    ACM Transactions on Graphics (ToG), 36(4), 1-13.

    Args:
        predictions: the predicted point cloud(s) with shape ([B,] NUM_POINTS, 3).
        groundtruth: the groundtruth point cloud(s) with shape ([B,] NUM_POINTS, 3).
        dist_pred_gt (optional): per point distances from predictions to groundtruth
            with shape ([B,] NUM_POINTS). Defaults to None.
        dist_gt_pred (optional): per point distances from groundtruth to predictions
            with shape ([B,] NUM_PTS_GT). Defaults to None.
        threshold: the distance treshold to use. Defaults to 0.01.

    Returns:
        A Tuple with:
        - The fscore with shape ([B,]).
        - The precision with shape ([B,]).
        - The recall with shape ([B,]).
    """
    pred_gt = predictions is not None and groundtruth is not None
    distances = dist_pred_gt is not None and dist_gt_pred is not None
    error_msg = (
        "Only one couple between (predictions, grountruths) and "
        "(dist_pred_gt, dist_gt_pred) should be not None"
    )
    assert (not pred_gt and distances) or (pred_gt and not distances), error_msg

    if predictions is not None and groundtruth is not None:
        pred_copy = torch.clone(predictions)
        gt_copy = torch.clone(groundtruth)

        batched, [pred_copy, gt_copy] = batchify([pred_copy, gt_copy], 3)

        dist_precision, _, _ = knn_points(pred_copy, gt_copy)
        dist_precision = dist_precision.squeeze(-1)
        dist_precision = sqrt(dist_precision)

        dist_recall, _, _ = knn_points(gt_copy, pred_copy)
        dist_recall = dist_recall.squeeze(-1)
        dist_recall = sqrt(dist_recall)
    else:
        dist_pred_gt_copy = torch.clone(dist_pred_gt)  # type: ignore
        dist_gt_pred_copy = torch.clone(dist_gt_pred)  # type: ignore

        batched, [dist_precision, dist_recall] = batchify([dist_pred_gt_copy, dist_gt_pred_copy], 2)

    batch_size = len(dist_precision)
    f_scores = zeros(batch_size).to(dist_precision.device)
    recall = zeros(batch_size).to(dist_precision.device)
    precision = zeros(batch_size).to(dist_precision.device)

    num_samples_prec = dist_precision.shape[1]
    num_samples_rec = dist_recall.shape[1]

    precision = sum((dist_precision < threshold), dim=-1) / max(num_samples_prec, 1)
    recall = sum((dist_recall < threshold), dim=-1) / max(num_samples_rec, 1)

    prec_and_rec = recall + precision
    mask = prec_and_rec > 0

    f_scores[mask] = 2 * recall[mask] * precision[mask] / prec_and_rec[mask]

    if batched:
        f_scores, precision, recall = unbatchify([f_scores, precision, recall])

    return f_scores, precision, recall



import torch
from torch import nn
from torch.autograd import Function

"""
Earth Mover's Distance for point clouds.

Code from:
https://github.com/Colin97/MSN-Point-Cloud-Completion

Author: Minghua Liu
"""

import emdcuda  # type: ignore


class EMD_CUDA_Function(Function):
    @staticmethod
    def forward(ctx, xyz1, xyz2, eps, iters):  # type: ignore

        batchsize, n, _ = xyz1.size()
        _, m, _ = xyz2.size()

        assert n == m
        assert xyz1.size()[0] == xyz2.size()[0]
        assert n % 1024 == 0
        assert batchsize <= 512

        xyz1 = xyz1.contiguous().float().cuda()
        xyz2 = xyz2.contiguous().float().cuda()
        dist = torch.zeros(batchsize, n, device="cuda").contiguous()
        assignment = torch.zeros(batchsize, n, device="cuda", dtype=torch.int32).contiguous() - 1
        assignment_inv = (
            torch.zeros(batchsize, m, device="cuda", dtype=torch.int32).contiguous() - 1
        )
        price = torch.zeros(batchsize, m, device="cuda").contiguous()
        bid = torch.zeros(batchsize, n, device="cuda", dtype=torch.int32).contiguous()
        bid_increments = torch.zeros(batchsize, n, device="cuda").contiguous()
        max_increments = torch.zeros(batchsize, m, device="cuda").contiguous()
        unass_idx = torch.zeros(batchsize * n, device="cuda", dtype=torch.int32).contiguous()
        max_idx = torch.zeros(batchsize * m, device="cuda", dtype=torch.int32).contiguous()
        unass_cnt = torch.zeros(512, dtype=torch.int32, device="cuda").contiguous()
        unass_cnt_sum = torch.zeros(512, dtype=torch.int32, device="cuda").contiguous()
        cnt_tmp = torch.zeros(512, dtype=torch.int32, device="cuda").contiguous()

        emdcuda.forward(
            xyz1,
            xyz2,
            dist,
            assignment,
            price,
            assignment_inv,
            bid,
            bid_increments,
            max_increments,
            unass_idx,
            unass_cnt,
            unass_cnt_sum,
            cnt_tmp,
            max_idx,
            eps,
            iters,
        )

        ctx.save_for_backward(xyz1, xyz2, assignment)
        return dist, assignment

    @staticmethod
    def backward(ctx, graddist, gradidx):  # type: ignore
        xyz1, xyz2, assignment = ctx.saved_tensors
        graddist = graddist.contiguous()

        gradxyz1 = torch.zeros(xyz1.size(), device="cuda").contiguous()
        gradxyz2 = torch.zeros(xyz2.size(), device="cuda").contiguous()

        emdcuda.backward(xyz1, xyz2, gradxyz1, graddist, assignment)
        return gradxyz1, gradxyz2, None, None


class EMD_CUDA(nn.Module):
    def __init__(self) -> None:
        super(EMD_CUDA, self).__init__()

    def forward(self, input1, input2, eps, iters):  # type: ignore
        return EMD_CUDA_Function.apply(input1, input2, eps, iters)



from pathlib import Path
from typing import Callable, List, Tuple

import h5py  # type: ignore
import open3d as o3d  # type: ignore
import torch
from torch import Tensor
from torch.utils.data import Dataset

from pycarus.geometry.pcd import get_o3d_pcd_from_tensor
from pycarus.transforms.var import Compose

T_ITEM = Tuple[Tensor, Tensor, Tensor]


class ShapeNetPartSegmentation(Dataset):
    """Class representing the ShapeNet part segmentation dataset as presented in:

    Yi, Li, et al. "A scalable active framework for region annotation in 3d shape collections."
    ACM Transactions on Graphics (ToG) 35.6 (2016): 1-12.
    """

    class_to_parts = {
        "02691156": [0, 1, 2, 3],
        "02773838": [4, 5],
        "02954340": [6, 7],
        "02958343": [8, 9, 10, 11],
        "03001627": [12, 13, 14, 15],
        "03261776": [16, 17, 18],
        "03467517": [19, 20, 21],
        "03624134": [22, 23],
        "03636649": [24, 25, 26, 27],
        "03642806": [28, 29],
        "03790512": [30, 31, 32, 33, 34, 35],
        "03797390": [36, 37],
        "03948459": [38, 39, 40],
        "04099429": [41, 42, 43],
        "04225987": [44, 45, 46],
        "04379243": [47, 48, 49],
    }

    part_colors = torch.tensor(
        [
            [0.65, 0.95, 0.05],
            [0.35, 0.05, 0.35],
            [0.65, 0.35, 0.65],
            [0.95, 0.95, 0.65],
            [0.95, 0.65, 0.05],
            [0.35, 0.05, 0.05],
            [0.65, 0.05, 0.05],
            [0.65, 0.35, 0.95],
            [0.05, 0.05, 0.65],
            [0.65, 0.05, 0.35],
            [0.05, 0.35, 0.35],
            [0.65, 0.65, 0.35],
            [0.35, 0.95, 0.05],
            [0.05, 0.35, 0.65],
            [0.95, 0.95, 0.35],
            [0.65, 0.65, 0.65],
            [0.95, 0.95, 0.05],
            [0.65, 0.35, 0.05],
            [0.35, 0.65, 0.05],
            [0.95, 0.65, 0.95],
            [0.95, 0.35, 0.65],
            [0.05, 0.65, 0.95],
            [0.65, 0.95, 0.65],
            [0.95, 0.35, 0.95],
            [0.05, 0.05, 0.95],
            [0.65, 0.05, 0.95],
            [0.65, 0.05, 0.65],
            [0.35, 0.35, 0.95],
            [0.95, 0.95, 0.95],
            [0.05, 0.05, 0.05],
            [0.05, 0.35, 0.95],
            [0.65, 0.95, 0.95],
            [0.95, 0.05, 0.05],
            [0.35, 0.95, 0.35],
            [0.05, 0.35, 0.05],
            [0.05, 0.65, 0.35],
            [0.05, 0.95, 0.05],
            [0.95, 0.65, 0.65],
            [0.35, 0.95, 0.95],
            [0.05, 0.95, 0.35],
            [0.95, 0.35, 0.05],
            [0.65, 0.35, 0.35],
            [0.35, 0.95, 0.65],
            [0.35, 0.35, 0.65],
            [0.65, 0.95, 0.35],
            [0.05, 0.95, 0.65],
            [0.65, 0.65, 0.95],
            [0.35, 0.05, 0.95],
            [0.35, 0.65, 0.95],
            [0.35, 0.05, 0.65],
        ]
    )

    def __init__(self, root: Path, split: str, transforms: List[Callable] = []) -> None:
        """Create an instance of ShapeNetPartSegmentation dataset.

        Args:
            root: The path to the root directory of the dataset.
            split: The split to use (only "train", "val" and "test" allowed).
            transforms: The transform to apply to the point clouds. Defaults to [].
        """
        super().__init__()

        assert split in ["train", "val", "test"]

        h5_files = [root / f for f in self.get_h5_files_from_split(split)]
        h5_files.sort()

        self.pcds, self.labels, self.part_labels = self.load_h5(h5_files)

        self.transform = Compose(transforms)

        with open(root / "all_object_categories.txt") as f:
            self.class_names = [line.strip().split("\t")[0] for line in f.readlines()]

    def get_h5_files_from_split(self, split: str) -> List[str]:
        """Get the h5 files to be loaded according to the given split.

        Args:
            split: The required split.

        Returns:
            The list of h5 files to be loaded.
        """
        files = {
            "train": [
                "ply_data_train0.h5",
                "ply_data_train1.h5",
                "ply_data_train2.h5",
                "ply_data_train3.h5",
                "ply_data_train4.h5",
                "ply_data_train5.h5",
            ],
            "val": ["ply_data_val0.h5"],
            "test": ["ply_data_test0.h5", "ply_data_test1.h5"],
        }
        return files[split]

    @staticmethod
    def load_h5(paths: List[Path]) -> Tuple[Tensor, Tensor, Tensor]:
        """Load all the given h5 files.

        Args:
            paths: The list of paths to the h5 files to load.

        Returns:
            - One tensor with all the pcds, with shape (NUM_PCDS, NUM_POINTS, 3).
            - One tensor with all the class labels, with shape (NUM_PCDS,).
            - One tensor with all the part labels, with shape (NUM_PCDS, NUM_POINTS,).
        """
        all_pcds: List[Tensor] = []
        all_labels: List[Tensor] = []
        all_part_labels: List[Tensor] = []

        for h5_path in paths:
            f = h5py.File(h5_path, "r+")
            pcds = f["data"][:].astype("float32")
            labels = f["label"][:].astype("int64")
            part_labels = f["pid"][:].astype("int64")
            f.close()
            all_pcds.append(torch.from_numpy(pcds))
            all_labels.append(torch.from_numpy(labels))
            all_part_labels.append(torch.from_numpy(part_labels))

        all_pcds_t = torch.cat(all_pcds, dim=0)
        all_labels_t = torch.cat(all_labels, dim=0).squeeze(-1)
        all_part_labels_t = torch.cat(all_part_labels, dim=0).squeeze(-1)

        return all_pcds_t, all_labels_t, all_part_labels_t

    def __len__(self) -> int:
        """Return the number of pcds in the dataset.

        Returns:
            The number of pcds in the dataset.
        """
        return self.pcds.shape[0]

    def __getitem__(self, index: int) -> T_ITEM:
        """Return the pcd, the class label and the part labels at the given index.

        Args:
            index: The index of the required element.

        Returns:
            - A tensor with the point cloud with shape (NUM_POINTS, 3).
            - A tensor with the label.
            - A tensor with the part labels with shape (NUM_POINTS,).
        """
        pcd = self.pcds[index]
        label = self.labels[index]
        part_labels = self.part_labels[index]

        pcd = self.transform(pcd)

        return pcd, label, part_labels

    @classmethod
    def color_pcd(cls, pcd: Tensor, part_labels: Tensor) -> o3d.geometry.PointCloud:
        """Prepare one item in order to be visualized using open3D draw geometries.

        Args:
            pcd: A tensor with the point cloud with shape (NUM_POINTS, 3).
            part_labels: A tensor with the labels for all the points with shape (NUM_POINTS,).

        Returns:
            The point cloud to draw with open 3D.
        """
        col = cls.part_colors[part_labels, :]

        return get_o3d_pcd_from_tensor(pcd, col)

    def get_class_name(self, label: int) -> str:
        """Get the class name from the class id.

        Args:
            label: The class numeric id.

        Returns:
            The class name.
        """
        return self.class_names[label]
